#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "bsh.h"

extern double cr2,cr3,cr4,cr6,cr12,cgp,c2gp,cgp_4,cr2gp,cgm0,crgm0;

extern int Na,Na1,Np,Np1;
extern double sa[];
extern double ss[],ssa[];
extern double sb[],sba[];
extern double gl[],gla[];
extern double sx[],sxa[];
extern double aS[],aSa[];
extern double aV[],aVa[],qaV[],qaVa[];
extern double gF[],gFa[],qgF[],qgFa[];
extern double gY[],gYa[],qgY[],qgYa[];
extern double BSH0[],BSK0[];

extern double aL[],aLa[];

extern double bsjs[],bsjp[];

static int RadC=0;
extern int ID;
extern int ESNa,ESNa1,ESNp,ESNp1,ESNpa;
extern double ESBt,ESRBt,ESRext;
extern double ESaR0,ESaZ0;

extern double ESgq[],EScs1[],ESsn1[];
extern double ESsa[],ESpa[];

extern double ESsr[],ESsra[],ESsrq[],ESsraq[];
extern double ESsz[],ESsza[],ESszq[],ESszaq[];
extern double ESaB[],ESaBa[],ESaBq[],ESaBaq[];
extern double ESgH[],ESgHa[],ESgHq[],ESgHaq[];
extern double BSHL[],BSHL1[],BSHL2[];

extern double ESgaG[];
extern double ESjs[],ESjb[];
extern double ESFF[],ESFF1a[];
extern double ESaP[],ESaP1a[];
extern double ESaT[],ESaT1a[];
extern double ESsp[],ESsp1a[];
extern double ESdgY[],ESdgY1a[];
extern double ESdgF[],ESdgF1a[];

static int nEq=6;
static double Y[16],dY[16];
static double RKy0[0x10],RKy1[0x10],RKy2[0x10],RKy3[0x10],RKy4[0x10];
static double rKy0[0x10],rKy1[0x10],rKy2[0x10],rKy3[0x10],rKy4[0x10];
static int K=4;
static double D,Da,A,X,dt,tend,pF0,F0=1.,g2F0=1.,g2l0=1.;

static double Ld[0x101];
static double a0=.9;

static int (*lRK)(double,double *,double *);
static int (*lrK)(double,double *,double *);

int rKStep(double *y, double *dy, double *a, double h,int neq)
{
  double x;
  int i;

  for(i=0; i < neq; i++){
    rKy1[i]	=h*dy[i];
    rKy0[i]	=y[i]+cr2*rKy1[i];
  }
  x	=(*a)+cr2*h;
  lrK(x,rKy0,rKy2);
  for(i=0; i < neq; i++){
    rKy2[i]	*=h;
    rKy0[i]	=y[i]+cr2*rKy2[i];
  }
  lrK(x,rKy0,rKy3);
  for(i=0; i < neq; i++){
    rKy3[i]	*=h;
    rKy0[i]	=y[i]+rKy3[i];
  }
  (*a)		+=h;
  lrK(*a,rKy0,rKy4);
  for(i=0; i < neq; i++){
    rKy4[i]	*=h;
    y[i]	+=cr6*(rKy1[i]+rKy4[i])+cr3*(rKy2[i]+rKy3[i]);
  }
  lrK(*a,y,dy);
  return(0);
}

int Y2dY(double s,double *y,double *dy)
{
  double r,t,y1,L;

  r	=1.-s;
  y1	=(2.-s)*s;
  X	=1.+y[1];
  dy[0]	=0.5*r*(y1+y[1]);
  dy[1]=-6.*y[0]*dy[0]/(y[2]+2.*X*y[0]);
  dy[2]=-5.*y[0]*dy[1];
  ESSetSplA(s/a0);
  splRA(&L,NULL,BSHL,BSHL2);
  dy[3]=-c2gp*L*dy[0];

  r	*=r;
  t	=12.*y[0];
  D	=sqrt(r*(4.*X-3.*r)+4.*t);
  y1	+=y[1];

  A	=(t+(3.*r-X)*y1)/(3.*r-2.*X+D);
  D	=(t-y1*y1)/(2.*X-r+D);

  dy[4]	=0.;
  t	=0.;
  lrK(t,dy+4,dy+6);
  while(t < tend) rKStep(dy+4,dy+6,&t,dt,1);
  dy[4]	*=sqrt(pF0-L*L*(y[2]-y[0]*X))*dy[0];
  dy[5]	=1.;
  return(0);
}

int Y2dYgF(double a,double *y,double *dy)
{
  double r,t,y1,Y1,gl,L;

  X	=1.+y[1];
  r	=1.-y[5];
  Y1	=r*r;
  y1	=y[1]+(2.-y[5])*y[5];
  gl	=12.*y[0];
  D	=sqrt(Y1*(X+3.*y1)+4.*gl);
  A	=(gl+(3.*Y1-X)*y1)/(Y1-2.*y1+D);
  D	=(gl-y1*y1)/(X+y1+D);

  ESSetSplA(y[5]/a0);
  splRA(&L,NULL,BSHL,BSHL2);

  dy[4]	=0.;
  t	=0.;
  lrK(t,dy+4,dy+6);
  while(t < tend) rKStep(dy+4,dy+6,&t,dt,1);

  dy[0]	=g2F0*a/(sqrt(pF0-L*L*(y[2]-y[0]*X))*dy[4]);
  dy[1]=-6.*y[0]*dy[0]/(y[2]+2.*X*y[0]);
  dy[2]=-5.*y[0]*dy[1];
  dy[3]=-c2gp*L*dy[0];
  dy[4]	=g2F0*a;
  dy[5]	=2.*dy[0]/(r*y1);

  return(0);
}

int Y2dYaV(double a,double *y,double *dy)
{
  double r,t,y1,Y1,gl,L;

  X	=1.+y[1];
  r	=1.-y[5];
  Y1	=r*r;
  y1	=y[1]+(2.-y[5])*y[5];
  gl	=6.*g2l0*a*a;
  D	=sqrt(Y1*(X+3.*y1)+4.*gl);
  A	=(gl+(3.*Y1-X)*y1)/(Y1-2.*y1+D);
  D	=(gl-y1*y1)/(X+y1+D);

  ESSetSplA(a);
  splRA(&L,NULL,BSHL,BSHL2);

  dy[4]	=0.;
  t	=0.;
  lrK(t,dy+4,dy+6);
  while(t < tend) rKStep(dy+4,dy+6,&t,dt,1);
  dy[0]	=g2l0*a;
  dy[1]=-6.*y[0]*dy[0]/(y[2]+2.*X*y[0]);
  dy[2]=-5.*y[0]*dy[1];
  dy[3]=-c2gp*L*dy[0];
  dy[4]	*=sqrt(pF0-L*L*(y[2]-y[0]*X))*dy[0];
  dy[5]	=2.*dy[0]/(r*y1);

  return(0);
}

int Z2dZ(double s,double *y,double *dy)
{
  dy[0]	=D+A*cos(s);
  dy[0]	*=0.5/sqrt(X+2.*D+dy[0]);
  return(0);
}

int BT2dgF(double t,double *y,double *dy)
{
  double Y;
  Y	=X+D+A*cos(t);
  if(Y <= 0.) Y =1.; 
  dy[0]	=1./(Y*sqrt(Y+2.*D));
  return(0);
}

int BT2HK0(double t,double *y,double *dy)
{
  double Y,s,cs;

  cs	=cos(t);
  Y	=X+D+A*cs;
  if(Y > 0.){
    s	=1./sqrt(Y+2.*D);
  }
  else{
    s	=1.;
    Y	=1.; 
  }
  dy[0]	=s/Y;
  dy[1]	=(A*(A+2.*D*cs)+D*D)*s;
  return(0);
}

int bsh2HK0()
{
  int i;
  double y[4],t;
  double Y1,y1,r;

  lRK	=BT2HK0;
  BSH0[0]=1.;
  BSK0[0]=8.;
  for(i=1; i < Na1; i++){
    X	=1.+sx[i];
    r	=1.-ss[i];
    Y1	=r*r;
    y1	=sx[i]+(2.-ss[i])*ss[i];
    t	=12.*gl[i];
    D	=sqrt(Y1*(X+3.*y1)+4.*t);
    A	=(t+(3.*Y1-X)*y1)/(Y1-2.*y1+D);
    D	=(t-y1*y1)/(X+y1+D);

    y[0]=0.;
    y[1]=0.;
    t	=0.;
    lRK(t,y,y+2);
    while(t < tend) RKStep(y,y+2,&t,dt,2);
    BSH0[i]=cr2gp*y[0];
    BSK0[i]=cr2gp*y[1]/gl[i];
  }
  return(0);
}

int BT2L0(double t,double *y,double *dy)
{
  double Y,dY,s,cs;

  cs	=cos(t);
  Y	=X+D+A*cs;
  if(Y > 0.){
    s	=sqrt(Y+2.*D);
    dY	=4.*(A*cs-D)/(A*A-D*D);
  }
  else{
    s	=1.;
    Y	=1.; 
    dY	=0.;
  }
  dy[0]	=1./(Y*s);
  dy[1]	=-(dY/Y+(0.5*dY+Da)/(Y+2.*D))*dy[0];
  return(0);
}

int RKStep(double *y, double *dy, double *a, double h,int neq)
{
  double x;
  int i;

  for(i=0; i < neq; i++){
    RKy1[i]	=h*dy[i];
    RKy0[i]	=y[i]+cr2*RKy1[i];
  }
  x	=(*a)+cr2*h;
  lRK(x,RKy0,RKy2);
  for(i=0; i < neq; i++){
    RKy2[i]	*=h;
    RKy0[i]	=y[i]+cr2*RKy2[i];
  }
  lRK(x,RKy0,RKy3);
  for(i=0; i < neq; i++){
    RKy3[i]	*=h;
    RKy0[i]	=y[i]+RKy3[i];
  }
  (*a)		+=h;
  lRK(*a,RKy0,RKy4);
  for(i=0; i < neq; i++){
    RKy4[i]	*=h;
    y[i]	+=cr6*(RKy1[i]+RKy4[i])+cr3*(RKy2[i]+RKy3[i]);
  }
  lRK(*a,y,dy);
  return(0);
}

int InitLd()
{
  int i;

  for(i=0; i < 0x101; i++) Ld[i]=1.;

  return(0);
}

int InitCharacteristics(double *y,double s,int iRC)
{
  int j;
  double g,b,c;

  ss[0]=0.;
  gl[0]=0.;
  sx[0]=0.;
  aS[0]=0.;
  gY[0]=0.;
  gF[0]=0.;
  aV[0]=0.;

  switch(iRC){
  case 0:
    gla[0]=a0*a0;
    sxa[0]=-3.*gla[0];
    aSa[0]=-gla[0];
    gYa[0]=-c2gp*BSHL[0]*gla[0];
    gFa[0]=c2gp*F0*gla[0];

    g	=0.5*s*s*(1.+(1.1875*s-1.5)*s);
    y[0]	=g;
    y[1]	=g*(1.125*g-3.);
    y[2]	=(7.5-3.75*g)*g*g;
    y[3]	=-c2gp*g*(BSHL[0]+0.5*BSHL2[0]*g/(a0*a0));
    y[4]	=c2gp*F0*g*(1+(3.5+0.25*BSHL[0]*BSHL[0]/pF0)*g);
    y[5]	=s;
    break;
  case 1:
    c	=gF[Na]/(cgp*F0);
    b	=7.+0.5*BSHL[0]*BSHL[0]/pF0;
    g	=c*s*s;
    g	/=1.+sqrt(1.+b*g);
    gla[0]=c;
    sxa[0]=-3.*gla[0];
    aSa[0]=-gla[0];
    gYa[0]=-c2gp*BSHL[0]*gla[0];
    gFa[0]=2.*gF[Na];
    ssa[0]=sqrt(c);
    y[0]	=g;
    y[1]	=g*(1.125*g-3.);
    y[2]	=(7.5-3.75*g)*g*g;
    y[3]	=-c2gp*g*(BSHL[0]+0.5*BSHL2[0]*g/(a0*a0));
    y[4]	=s*s*gF[Na];
    y[5]	=sqrt(2.*g);
    y[5]	=sqrt(2.*g/(1.-1.5*y[5]));
    y[5]	=sqrt(2.*g/(1.+(1.1875*y[5]-1.5)*y[5]));
    break;
  case 2:
    gla[0]	=g2l0;
    sxa[0]	=-3.*gla[0];
    aSa[0]	=-gla[0];
    gYa[0]	=-c2gp*BSHL[0]*gla[0];
    gFa[0]	=c2gp*F0*gla[0];
    ssa[0]	=sqrt(g2l0);

    g		=gl[Na]*s*s;
    y[0]	=g;
    y[1]	=g*(1.125*g-3.);
    y[2]	=(7.5-3.75*g)*g*g;
    y[3]	=-c2gp*g*(BSHL[0]+cgp*cgp*BSHL2[0]*g/aV[Na]);
    y[4]	=c2gp*F0*g*(1+(3.5+0.25*BSHL[0]*BSHL[0]/pF0)*g);
    y[5]	=sqrt(2.*g);
    y[5]	=sqrt(2.*g/(1.-1.5*y[5]));
    y[5]	=sqrt(2.*g/(1.+(1.1875*y[5]-1.5)*y[5]));
    break;
  }
  aVa[0]=c2gp*c2gp*gla[0];
  return(0);
}

int RadialCoordinates()
{
  int i,k;
  double s,V0,gF0,gY0;
  double r,Y1,y1,t,tt;

  gY0	=1./sqrt(-gY[ESNa]);
  gF0	=1./sqrt(gF[ESNa]);
  V0	=1./sqrt(aV[ESNa]);
  i	=0;
  sb[i]		=0.;
  qgY[i]	=0.;
  qgF[i]	=0.;
  qaV[i]	=0.;
  sba[i]	=a0;
  qgYa[i]	=sqrt(cgp*BSHL[0])*gY0*a0;
  qgFa[i]	=sqrt(cgp*F0)*gF0*a0;
  qaVa[i]	=cgp*sqrt(2.)*V0*a0;
  for(i=1; i < ESNa1; i++){
    X	=1.+sx[i];
    r	=1.-ss[i];
    Y1	=r*r;
    y1	=sx[i]+(2.-ss[i])*ss[i];
    t	=12.*gl[i];
    D	=sqrt(Y1*(X+3.*y1)+4.*t);
    A	=(t+(3.*Y1-X)*y1)/(Y1-2.*y1+D);
    D	=(t-y1*y1)/(X+y1+D);
    tt	=acos(-D/A)-dt;
    t	=0.;
    sb[i]=0.;
    Z2dZ(t,sb+i,sba+i);
    while(t < tt){
      for(k=0; k < K; k++) RKStep(sb+i,sba+i,&t,dt,1);
    }
    RKStep(sb+i,sba+i,&t,tt+dt-t,1);

    sba[i]	=2.*gla[i]*a0/(sqrt((A*A-D*D)*(X+2.*D)));
    s	=sqrt(-gY[i]);
    qgY[i]	=s*gY0;
    qgYa[i]	=0.5*gYa[i]*gY0/s;
    s	=sqrt(gF[i]);
    qgF[i]	=s*gF0;
    qgFa[i]	=0.5*gFa[i]*gF0/s;
    s	=sqrt(aV[i]);
    qaV[i]	=s*V0;
    qaVa[i]	=0.5*aVa[i]*V0/s;
  }
  return(0);
}

int bsh2RZ()
{
  int i,ii,j,ji,k;
  double r,Y1,t,u[2],v[2],y1;
  double y[4];

  lRK	=Z2dZ;

  ji	=0;
  for(j=0; j < Np1; j++){
    ESsr[ji]	=1.;
    ESsz[ji]	=0.;
    ji++;
  }
  for(i=1; i < Na1; i++){
    X	=1.+sx[i];
    r	=1.-ss[i];
    Y1	=r*r;
    y1	=sx[i]+(2.-ss[i])*ss[i];
    t	=12.*gl[i];

    D	=sqrt(Y1*(X+3.*y1)+4.*t);
    A	=(t+(3.*Y1-X)*y1)/(Y1-2.*y1+D);
    D	=(t-y1*y1)/(X+y1+D);
    ii	=0;
    u[0]	=sx[i];
    u[1]	=sx[i]+1e-6;
    do{
      zero(&ii,u,v,sx+i,y[0]);
      X	=1.+sx[i];
      t	=0.;
      y[0]=0.;
      Z2dZ(t,y,y+2);
      for(j=1; j < Np1; j++){
	for(k=0; k < K; k++) RKStep(y,y+2,&t,dt,1);
      }
#ifdef H
      out("??+?? ii d",i,ii,y[0]);
#endif    
    }while(ii > 0 && ii < 15 && fabs(y[0]) > 1e-12);
    t	=0.;
    y[0]=0.;
    Z2dZ(t,y,y+2);
    ESsr[ji]	=sqrt(X+D+A*EScs1[0]);
    ESsz[ji]	=y[0];
    ji++;
    if(ss[i] < 1.){
      for(j=1; j < Np; j++){
	for(k=0; k < K; k++) RKStep(y,y+2,&t,dt,1);
	ESsr[ji]	=sqrt(X+D+A*EScs1[j]);
	ESsz[ji]	=-y[0];
	ji++;
      }
    }
    else{
      for(j=1; j < Np; j++){
	for(k=0; k < K; k++) RKStep(y,y+2,&t,dt,1);
	ESsr[ji]	=sqrt(fabs(X+D+A*EScs1[j]));
	ESsz[ji]	=-y[0];
	ji++;
      }
    }
    j	=ji-Np;
    ESsr[ji]	=ESsr[j];
    ESsz[ji]	=ESsz[j];
    ji++;
  }
  return(0);
}

int bsh2basisP()
{
  int i;
  double dgl,L,La,Laa,F,Fa;
  double s,r,Y1,y1,t;
  double y[8];
  double L0,L0a;

  i	=0;
  L	=BSHL[i];
  ESdgY[i]	=cr2gp*gYa[0];
  ESdgF[i]	=cr2gp*gFa[0];

  ESaT[i]	=-0.5*L;
  ESaP[i]	=2.5*L;

  ESdgY1a[i]	=0.;
  ESdgF1a[i]	=0.;
  ESaT1a[i]	=0.;
  ESaP1a[i]	=0.;
  aL[i]		=1.;
  aLa[i]	=7.*g2l0;

  for(i=1; i < Na1; i++){
    X	=1.+sx[i];
    r	=1.-ss[i];
    Y1	=r*r;
    y1	=sx[i]+(2.-ss[i])*ss[i];
    t	=12.*gl[i];
    D	=sqrt(Y1*(X+3.*y1)+4.*t);
    A	=(t+(3.*Y1-X)*y1)/(Y1-2.*y1+D);
    D	=(t-y1*y1)/(X+y1+D);
    Da	=2./(X+2.*D);
    dgl	=g2l0*ESsa[i];

    lRK	=BT2L0;
    y[0]=0.;
    y[1]=0.;
    t	=0.;
    lRK(t,y,y+2);
    while(t < tend) RKStep(y,y+2,&t,dt,2);

    L0	=cr2gp*y[0];
    L0a	=cr2gp*y[1];

    L0a	=cr2gp*y[1]*dgl;
    aL[i]	=L0;
    aLa[i]	=L0a;

    L	=BSHL[i];
    La	=BSHL1[i];
    Laa	=BSHL2[i];

    ESdgY[i]	=-L*gla[i];
    ESdgY1a[i]	=-La*gla[i];

    F	=sqrt(pF0-aS[i]*L*L);
    Fa	=(aS[i]*La+0.5*aSa[i]*ESsa[i]*L)*L/F;

    ESdgF[i]	=F*L0*gla[i];
    ESdgF1a[i]	=(Fa*L0+F*L0a)*gla[i];

    ESaP[i]	=2.5*L+1.5*ESsa[i]*La;
    ESaP1a[i]	=4.*La+1.5*ESsa[i]*Laa;
 
    y1	=1./(3.*X*gl[i]+aS[i]);
    F	=18.*gl[i]*gl[i]*y1;
    s	=F-0.5*X;

    ESaT1a[i]	=aS[i]*La/dgl;
    ESaT[i]	=ESaT1a[i]+s*L;

    ESaT1a[i]	=3.*s*La+aS[i]*Laa/dgl-ESaT1a[i]/ESsa[i]
      +(39*gl[i]-F*(F+2.*X))*L*y1*dgl;
  }
  ESRBt	=F;
  ESRext=1.;
  ESBt	=F;
  return(0);
}

int bsh2basis2D()
{
  int i,ii,j,ji,k;
  double rr,s,ad,cs,sn,Y1,t,u[2],v[2],y1;
  double y[4];
  double r,ra,rq,raq,Bt,Bta,Btq,Btaq,L0,L0a,L,La,Lq,Laq,dgl;

  ji	=0;
  ad	=sqrt(g2l0);
  for(j=0; j < Np1; j++){
    cs	=EScs1[j];
    sn	=ESsn1[j];
    ra	=cs*ad;
    raq	=-sn*ad;

    ESsr[ji]	=1;
    ESsra[ji]	=ra;
    ESsrq[ji]	=0.;
    ESsraq[ji]	=raq;

    ESsz[ji]	=0.;
    ESsza[ji]	=-ad*sn;
    ESszq[ji]	=0.;
    ESszaq[ji]	=-ad*cs;

    ESaB[ji]	=F0;
    ESaBa[ji]	=-F0*ra;
    ESaBq[ji]	=0.;
    ESaBaq[ji]	=-F0*raq;

    ESgH[ji]	=0.;
    ESgHa[ji]	=-3.*ra;
    ESgHq[ji]	=0.;
    ESgHaq[ji]	=-3.*raq;

    ji++;
  }

  lRK	=Z2dZ;
  for(i=1; i < Na1; i++){
    dgl	=g2l0*ESsa[i];
    X	=1.+sx[i];
    r	=1.-ss[i];
    Y1	=r*r;
    y1	=sx[i]+(2.-ss[i])*ss[i];
    t	=12.*gl[i];
    D	=sqrt(Y1*(X+3.*y1)+4.*t);
    A	=(t+(3.*Y1-X)*y1)/(Y1-2.*y1+D);
    D	=(t-y1*y1)/(X+y1+D);
    ad	=A > D ? dgl/(A*A-D*D) : 1.;

    ii	=0;
    u[0]	=sx[i];
    u[1]	=sx[i]+1e-6;
    do{
      zero(&ii,u,v,sx+i,y[0]);
      X	=1.+sx[i];
      t	=0.;
      y[0]=0.;
      Z2dZ(t,y,y+2);
      for(j=1; j < Np1; j++){
	for(k=0; k < K; k++) RKStep(y,y+2,&t,dt,1);
      }
    }while(ii > 0 && ii < 15 && fabs(y[0]) > 1e-12);
    L0	=1./aL[i];
    L0a	=aLa[i]*dgl*L0;
    Da	=2.*dgl/(X+2.*D);

    t	=0.;
    y[0]=0.;
    Z2dZ(t,y,y+2);
    for(j=0; j < Np; j++){
      cs	=EScs1[j];
      sn	=ESsn1[j];
      r		=sqrt(X+D+A*cs);
      rr	=1./r;
      ra	=2.*(A*cs-D)*ad*rr;
      rq	=-0.5*A*sn*rr;
      raq	=-(rq*ra+2.*A*sn*ad)*rr;
      ESsr[ji]	=r;
      ESsra[ji]	=ra;
      ESsrq[ji]	=rq;
      ESsraq[ji]=raq;

      Bt	=pF0*rr*rr;
      Bta	=-2.*Bt*ra*rr;
      Btq	=-2.*Bt*rq*rr;
      Btaq	=-(3.*Bta*rq+2.*Bt*raq)*rr;

      L	=BSHL[i];
      Bt	+=3.*gl[i]*L*L;
      Bta	+=3.*(dgl*L+2.*gl[i]*BSHL1[i])*L;
      Bt	=sqrt(Bt);

      ESaB[ji]	=Bt;
      ESaBa[ji]	=0.5*Bta/Bt;
      ESaBq[ji]	=0.5*Btq/Bt;
      ESaBaq[ji]=(0.5*Btaq-ESaBa[ji]*ESaBq[ji])/Bt;
    
      s		=1./(r*r+2.*D);
      Bt	=sqrt(s);
      Bta	=-(r*ra+Da)*Bt*s;
      Btq	=-r*rq*Bt*s;
      Btaq	=-(ra*rq*Bt+r*(raq*Bt+3.*rq*Bta))*s;

      L		=rr*rr*L0;
      La	=-(2.*ra*rr+L0a)*L;
      Lq	=-2.*rq*rr*L;
      Laq	=-(2.*raq*L+3.*Lq*ra)*rr-Lq*L0a;

      ESgH[ji]	=L*Bt-1.;
      ESgHa[ji]	=La*Bt+L*Bta;
      ESgHq[ji]	=Lq*Bt+L*Btq;
      ESgHaq[ji]=Laq*Bt+Lq*Bta+La*Btq+L*Btaq;

      ESsz[ji]	=-y[0];
      ESsza[ji]	=-2.*ad*A*sn*Bt;
      ESszq[ji]	=-0.5*(A*cs+D)*Bt;
      ESszaq[ji]	=-(2.*A*ad*cs+0.5*Da)*Bt+0.5*(A*cs+D)*Bta;

      for(k=0; k < K; k++) RKStep(y,y+2,&t,dt,1);
      ji++;
    }
    j	=ji-Np;
    ESsr[ji]	=ESsr[j];
    ESsra[ji]	=ESsra[j];
    ESsrq[ji]	=ESsrq[j];
    ESsraq[ji]	=ESsraq[j];
    
    ESaB[ji]	=ESaB[j];
    ESaBa[ji]	=ESaBa[j];
    ESaBq[ji]	=ESaBq[j];
    ESaBaq[ji]	=ESaBaq[j];
    
    ESgH[ji]	=ESgH[j];
    ESgHa[ji]	=ESgHa[j];
    ESgHq[ji]	=ESgHq[j];
    ESgHaq[ji]	=ESgHaq[j];
    
    ESsz[ji]	=ESsz[j];
    ESsza[ji]	=ESsza[j];
    ESszq[ji]	=ESszq[j];
    ESszaq[ji]	=ESszaq[j];
    ji++;
  }
  return(0);
}

int bshPlotRZ(int iWin)
{
  double x[2],y[2];
  x[0]	=0.;
  x[1]	=1.5;
  y[0]	=-.5;
  y[1]	=.5;
  PSFileOpen(1,1);

  SetPlotName("r","z","Bishop-Taylor");
  Scale2d(iWin,x,y,2,6,0);
  Plot2d(iWin,ESsr,ESsz,Np1*Na1,6,0,0,0);

  PSNewFrame();
  PSPlot2d(ESsr,ESsz,Np1*Na1,6,0,0,0);
  PSFileClose();

  return(0);
}

int EquilBTAO()
{
  FILE *lF;
  int i,j,ji,m;
  char *lD,*lN;
  char FNm[0x100];
  double g[0x100],g2[0x100],x[0x100],s[0x100];
  double a,f,L,dL;
  
  CbGetCurrentCbName(&lD,&lN);
  m	=sprintf(FNm,"%s/Out/PlVac.ai",lD)-8;
  lF	=fopen(FNm,"w");

  L	=BSHL[Na];
  fputs("<PlVac RBtor>[4](%e)\n",lF);
  fprintf(lF,"%8.5f -vacuum R x Btor toroidal field and its reference radius\n"
	  ,sqrt(pF0-L*L*aS[Na]));
  fprintf(lF,"%8.5f -reference radius of the vacuum magmetic field\n",1.);
  fprintf(lF,"%8.5f - R0 - approximate R and Z of the magnetic axis\n",1.);
  fprintf(lF,"%8.5f - Z0\n",0.);
  fprintf(lF,"\n<PlVac Boundary>[%d]%s\n",Np1,"(%d %e %e)");

  ji	=Np1*Na;
  for(j=0,ji=Np1*Na; j < Np1; j++,ji++)
    fprintf(lF,"%3d %12.5e %12.5e\n",j,ESsr[ji],ESsz[ji]);
  fclose(lF);
  printf("Output: %s\n",FNm);

  sprintf(FNm+m,"PlCr.ai",lD);
  lF	=fopen(FNm,"w");
  fprintf(lF,"%2d - j_s\n",Na1);
  for(i=0; i < Na1; i++){
    g[i]	=bsjs[i];
    x[i]	=sqrt(gF[i]/gF[Na]);
  }
  spl(g,g2,x,NULL,x,&Na);
  for(i=0; i < Na1; i++){
    a	=ss[i]/ss[Na];
    splr0(&f,NULL,&a,g,g2,x,&Na);
    fprintf(lF,"%2d %12.5e %12.5e\n",i,ss[i]/ss[Na],f);
  }
  fclose(lF);
  printf("Output: %s\n",FNm);

  sprintf(FNm+m,"PlPr.ai",lD);

  lF	=fopen(FNm,"w");
  fprintf(lF,"%2d - j_p\n",Na1);
  for(i=0; i < Na1; i++) g[i]=bsjp[i];
  spl(g,g2,x,NULL,x,&Na);
  for(i=0; i < Na1; i++){
    a	=ss[i]/ss[Na];
    splr0(&f,NULL,&a,g,g2,x,&Na);
    fprintf(lF,"%2d %12.5e %12.5e\n",i,ss[i]/ss[Na],f);
  }
  fclose(lF);
  printf("Output: %s\n",FNm);

  return(0);
}

int BTApprox()
{
  int i;
  double s;
  double s0,s1,ds0,ds1;
  double h,t,T,f,h0;


  h	=gl[Na];

  s0	=-3.;
  ds0	=1.125*h;
  s1	=sx[Na]/gl[Na];
  ds1	=(sxa[Na]/(gla[Na]*gl[Na])-s1/gl[Na])*h;

  for(i=1; i < Na1; i++){
    t	=gl[i]/h;
    T	=1.-t;
    s	=gl[i]*(T*T*(3.-2.*T)*s0+t*t*(3.-2.*t)*s1+t*T*(T*ds0-t*ds1));
    printf("sx[%2d] %14.6e %14.6e %14.6e %10.3e\n"
	   ,i,sx[i],s,sx[i]/s,sx[i]/s-1.);
  }
  printf("sx: %14.6e %14.6e %14.6e %14.6e\n",s0,ds1,s1,ds1);
  printf("gl: %17.9e\n",gl[Na]);

  h0=9.931550123e-02;

  s0	=-1./h0;
  ds0	=10.5-1./h0;

  f	=1./(gl[Na]*(h0-gl[Na]));
  X	=1.+sx[Na];
  ds1	=gl[Na]/X;
  s1	=X-12.*ds1;
  ds1	=(1.5*sxa[Na]/gla[Na]+(6.*sxa[Na]/(gla[Na]*X)+12./X/X)*ds1 -6./X)*h0;

#ifdef H
  aS[i]	=10.5*g*g-g;
#endif

  for(i=1; i < Na; i++){
    t	=gl[i]/h;
    T	=1.-t;
    s	=gl[i]*(h0-gl[i])*(T*T*(3.-2.*T)*s0+t*t*(3.-2.*t)*s1+t*T*(T*ds0-t*ds1));
    printf("aS[%2d] %14.6e %14.6e %14.6e %10.3e\n"
	   ,i,aS[i],s,aS[i]/s,aS[i]/s-1.);
  }
  printf("aS: %14.6e %14.6e %14.6e %14.6e\n",s0,ds1,s1,ds1);

#ifdef H
  h=gl[Na];
  s0	=-1.;
  ds0	=10.5*h;

  f	=1./gl[Na];
  s1	=aS[Na]*f;
  ds1	=aSa[Na]*f*h/gla[Na]-s1*f*h;
#ifdef H
  aS[i]	=10.5*g*g-g;
#endif

  for(i=1; i < Na1; i++){
    t	=gl[i]/h;
    T	=1.-t;
    s	=gl[i]*(T*T*(3.-2.*T)*s0+t*t*(3.-2.*t)*s1+t*T*(T*ds0-t*ds1));
    printf("aS[%2d] %14.6e %14.6e %14.6e %10.3e\n"
	   ,i,aS[i],s,aS[i]/s,aS[i]/s-1.);
  }
  printf("aS: %14.6e %14.6e %14.6e %14.6e\n",s0,ds1,s1,ds1);
#endif
  return(0);
}



int EquilBT()
{
  static int Fla=0,Fld=0,Fl=0;
  int id[4]={909,902,901,0};
  
  int i,ii,j,ji,k,n;
  double s,h,r,y1,Y1,t;
  double y[4],*lP,*ld,*lx;
  double u[2],v[2];
  
  lrK	=BT2dgF;

#ifndef Tbl_EquilBT
  ID	=id[RadC];
  if(Na != ESNa || Np != ESNp){
    InitBsh(Na,Np);
    InitLd();
  }
  Na1	=ESNa1;
  Np1	=ESNp1;
  s	=1./ESNa;
  for(i=0; i < Na1; i++) ESsa[i]=s*i;
  for(i=0; i < Na1; i++) BSHL[i]=Ld[i];

#ifdef H
  f2splA(BSHL,BSHL2,ESsa,NULL,Ld);
#endif
  f2spl(BSHL,BSHL2,ESsa,NULL,Ld,NULL,ESsa,&ESNa,ESgaG,ESgaG+1,ESgaG+2,ESgaG+3);
  splA1a(BSHL,BSHL1,BSHL2);

  pF0	=F0*F0;
  if(a0 > 1.) a0=1.;
  if(a0 < 0.) a0=0.5;
  h	=a0/(Na*K);
  dt	=c2gp/(Np*K);
  tend	=c2gp-0.5*dt;

  lRK	=Y2dY;
  s	=h;
  InitCharacteristics(Y,s,0);

  bsjp[0]=2.5*BSHL[0];
  bsjs[0]=bsjp[0]-0.5*BSHL[0];
  
  lRK(s,Y,dY);
  for(k=1; k < K; k++) RKStep(Y,dY,&s,h,nEq);
  for(i=1; i < Na; i++){
    ss[i]	=K*h*i;
    gl[i]	=Y[0];
    sx[i]	=Y[1];
    aS[i]	=Y[2]-Y[0]*(1.+Y[1]);
    gY[i]	=Y[3];
    gF[i]	=Y[4];
    aV[i]	=c2gp*c2gp*gl[i];

    gla[i]	=dY[0]*a0/ESsa[i];
    sxa[i]	=dY[1]*a0/ESsa[i];
    aSa[i]	=(dY[2]-dY[0]*(1.+Y[1])-Y[0]*dY[1])*a0/ESsa[i];
    gYa[i]	=dY[3]*a0/ESsa[i];
    gFa[i]	=dY[4]*a0/ESsa[i];
    aVa[i]	=c2gp*c2gp*gla[i];

    bsjp[i]	=2.5*BSHL[i];
    bsjs[i]	=bsjp[i]+0.5*BSHL[i]*((dY[2]-Y[0]*dY[1])/dY[0]-1.-Y[1]);
    for(k=0; k < K; k++) RKStep(Y,dY,&s,h,nEq);
  }
  ss[i]	=a0;
  gl[i]	=Y[0];
  sx[i]	=Y[1];
  aS[i]	=Y[2]-Y[0]*(1.+Y[1]);
  gY[i]	=Y[3];
  gF[i]	=Y[4];
  aV[i]	=c2gp*c2gp*gl[i];
  gla[i]=dY[0]*a0/ESsa[i];
  sxa[i]=dY[1]*a0/ESsa[i];
  aSa[i]=(dY[2]-dY[0]*(1.+Y[1])-Y[0]*dY[1])*a0/ESsa[i];
  gYa[i]=dY[3]*a0/ESsa[i];
  gFa[i]=dY[4]*a0/ESsa[i];
  aVa[i]=c2gp*c2gp*gla[i];

  BTApprox();

  bsjp[i]=2.5*BSHL[i];
  bsjs[i]	=dY[0] == 0. ? bsjs[i-1]
    : bsjp[i]+0.5*BSHL[i]*((dY[2]-Y[0]*dY[1])/dY[0]-1.-Y[1]);
  if(RadC){
    g2l0	=2.*gl[Na];
    g2F0	=2.*gF[Na];
    lRK	=RadC == 1 ? Y2dYgF : Y2dYaV;
    h	=1./(Na*K);
    s		=h;
    InitCharacteristics(Y,s,RadC);

    bsjp[0]	=2.5*BSHL[0];
    bsjs[0]	=bsjp[0]-0.5*BSHL[0];

    lRK(s,Y,dY);
    for(k=1; k < K; k++) RKStep(Y,dY,&s,h,nEq);
    for(i=1; i < Na; i++){
      sx[i]	=Y[1];
      aS[i]	=Y[2]-Y[0]*(1.+Y[1]);
      gY[i]	=Y[3];
      ss[i]	=Y[5];
      sxa[i]	=dY[1]/ESsa[i];
      aSa[i]	=(dY[2]-dY[0]*(1.+Y[1])-Y[0]*dY[1])/ESsa[i];
      gYa[i]	=dY[3]/ESsa[i];
      ssa[i]	=dY[5];
      if(RadC == 1){
	gl[i]	=Y[0];
	gF[i]	=ESsa[i]*ESsa[i]*gF[Na];
	gla[i]	=dY[0]/ESsa[i];
	gFa[i]	=g2F0;
      }
      else{
	gl[i]	=ESsa[i]*ESsa[i]*gl[Na];
	gF[i]	=Y[4];
	gla[i]	=g2l0;
	gFa[i]	=dY[4]/ESsa[i];
      }
      aV[i]	=c2gp*c2gp*gl[i];
      aVa[i]	=c2gp*c2gp*gla[i];

      bsjp[i]	=2.5*BSHL[i];
      bsjs[i]	=bsjp[i]+0.5*BSHL[i]*((dY[2]-Y[0]*dY[1])/dY[0]-1.-Y[1]);
      for(k=0; k < K; k++) RKStep(Y,dY,&s,h,nEq);
    }

    sx[i]	=Y[1];
    aS[i]	=Y[2]-Y[0]*(1.+Y[1]);
    gY[i]	=Y[3];
    ss[i]	=Y[5];

    sxa[i]	=dY[1]/ESsa[i];
    aSa[i]	=(dY[2]-dY[0]*(1.+Y[1])-Y[0]*dY[1])/ESsa[i];
    gYa[i]	=dY[3]/ESsa[i];
    ssa[i]	=dY[5];
    if(RadC == 1){
      gl[i]	=Y[0];
      gla[i]	=dY[0]/ESsa[i];
      gFa[i]	=g2F0;
    }
    else{
      gF[i]	=Y[4];
      gla[i]	=g2l0;
      gFa[i]	=dY[4]/ESsa[i];
    }
    aV[i]	=c2gp*c2gp*gl[i];
    aVa[i]	=c2gp*c2gp*gla[i];
    bsjp[i]	=2.5*BSHL[i];
    bsjs[i]	=dY[0] == 0. ? bsjs[i-1]
      : bsjp[i]+0.5*BSHL[i]*((dY[2]-Y[0]*dY[1])/dY[0]-1.-Y[1]);
  }
  ESaR0=1.;
  ESaZ0=0.;
  lRK	=Z2dZ;
  RadialCoordinates();
  bsh2RZ();
  bsh2HK0();
  if(a0 < 1. && RadC == 2){
    bsh2basisP();
    bsh2basis2D();
    CbStr2UserMessage("ESI data are ready");
  }
  switch(Fla){
  case 0: 
    lx	=ESsa;
    break;
  case 1:
    lx	=sb;
    break;
  case 2:
    lx	=qaV;
    break;
  case 3:
    lx	=qgF;
    break;
  case 4:
    lx	=qgY;
    break;
  }
  lP	=NULL;

  lx	=gl;
  switch(Fl){
  case 0: 
    ld	=Ld;
    lP	=BSHL;
    break;
  case 1: 
    ld	=Fld ? gla : gl;
    break;
  case 2:
    ld	=Fld ? sxa : sx;
    break;
  case 3:
    ld	=Fld ? aSa : aS;
    break;
  case 4:
    ld	=Fld ? gFa : gF;
    break;
  case 5:
    ld	=Fld ? gYa : gY;
    break;
  case 6:
    ld	=Fld ? aVa : aV;
    break;
  case 7:
    ld	=BSH0;
    break;
  case 8:
    ld	=BSK0;
    break;
  default:
    ld	=gl;
  }
  if(lP == NULL) lP=ld;
  n	=Na1;
  bshPlotRZ(2);

  printf(" '~'=%d\n",'~');

#endif /*Tbl_EquilBT*/

#ifndef AscOEquilBT
#undef AscOEquilBT
#endif/*AscOEquilBT*/
  return(0);
}
