#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <sys/shm.h>

#include "esi.h"

#ifndef single
typedef double REAL;
#else
typedef float REAL;
#endif

#ifndef mzl_ESI

/***************** Names of ESI groups of variables ************/
char GrName[16][32]={
  "<ESI dimensions>",
  "<ESI [gq]>",
  "<ESI [sa]>",
  "<ESI [baF bsp gF gY]>",
  "<ESI bgFa bgFaa bgYa bgYaa>",
  "<ESI T Ta P Pa>",
  "<ESI r ra rq raq>",
  "<ESI z za zq zaq>",
  "<ESI B Ba Bq Baq>",
  "<ESI ghq ghaq ghqq ghaqq>",
  "<ESI [gz]>",   		/* only for 3D */
  "<ESI rgz ragz rqgz>",   	/* only for 3D */
  "<ESI zgz zagz zqgz>",   	/* only for 3D */
  "<ESI Bgz Bagz Bqgz>",   	/* only for 3D */
  "<ESI ghgz ghagz ghqgz>",	/* only for 3D */
  ""
};
/***************** ESI variables ****************************/
static double cr2,cr3,cr4,cr5,cr6,cr7,cr8,cr12,cgp,c2gp,cgp_4,cr2gp,cgm0,crgm0;
static double cbsp2ne;
static int kD=8,kI=4;

static int nESI=4;
static ESI sESI[4];

static int Na,Na1=0,Np,Np1=0,Npa;
static int ID=999999;

static REAL RBtor=1.;
static REAL aRBtor=1.;
static REAL Rext=1.;
static REAL *gq=NULL,*sa;
static REAL *dgF,*d2gF,*dgY,*d2gY,*aT,*daT,*aP,*daP;
static REAL *gF,*gY;
static REAL *bsp,*dbsp;
static REAL *aF,*daF;

static REAL *sr,*sra,*srq,*sraq;
static REAL *sz,*sza,*szq,*szaq;
static REAL *aB,*aBa,*aBq,*aBaq;
static REAL *gH,*gHa,*gHq,*gHaq,*gHqq,*gHaqq;

static int i0=0,i1=1,j00,j10,j01,j11; 
static REAL A,ha,rha;
static REAL hq,rhq,cgq0,crgq0;/* period and inverse period in gq */
static REAL xF,XF,xD,XD,dX,dxD,dXD,yF,YF,yD,YD,dY,dyD,dYD;
static REAL y,Y,yy,YY;

/***************** Pointers to user space ****************************/
static int *sw;
static REAL *r=NULL,*ra=NULL,*rq=NULL,*z=NULL,*za=NULL,*zq=NULL;
static REAL *B=NULL,*Ba=NULL,*Bq=NULL,*gh=NULL,*gha=NULL,*ghq=NULL;
static REAL *gFa=NULL,*gFaa=NULL,*gYa=NULL,*gYaa=NULL;
static REAL *T=NULL,*Ta=NULL,*P=NULL,*Pa=NULL;

static REAL *F=NULL,*Fa=NULL;
static REAL *sp=NULL,*spa=NULL;

static double cme=9.1094e-28; /* [g] - electron mass */
static double cmi=1.6726e-24; /* [g] - proton mass */
static double cc=2.9979e+10; /* [cm/sec] - speed of light */
static double ce=4.8032e-10; /* [CGS] - proton electric charge */
static double cEeV=1.6022e-12; /* [erg] energy of 1 eV */
static double cEkeV=1.6022e-9; /* [erg] energy of 1 keV */

/***************** Hermit polynomials  ****************************/
#ifndef stg_Hermit
int PHermit(double *f, double *df,double x,double *g,double *dg)
{
  double t;
  double p1,p2,p3;

  p1	=ha*dg[0];
  p3	=ha*dg[1];
  t	=g[1]-g[0];
  p2	=3.*t-2.*p1-p3;
  p3	+=p1-2.*t;

  t	=x*rha;

  *f	=g[0]+(p1+(p2+p3*t)*t)*t;
  *df	=rha*(p1+(2.*p2+3.*p3*t)*t);
  return(0);
}

int EPPAHermit(double *f, double *df,double x,int i)
{
  int i0,i1;
  double t,pt;
  double y0,y1,y2,y3,y4;
  double q,q0,q1,q2,q3;
  double r0,r1,r2,r3;

  i0	=i;
  i1	=i0+1;
  t	=x*rha;
  if(t == 0.){
    *f	=0.;
    *df	=2.*aT[i0]*dgY[i0]*sa[i0];
    return(0);
  }
  pt	=t*t;

  y0	=sa[i0]*dgY[i0];
  y1	=dgY[i0];
  y2	=ha*d2gY[i0];
  y4	=ha*d2gY[i1];
  q0	=dgY[i1]-y1;
  y3	=3.*q0-2.*y2-y4;
  y4	+=y2-2.*q0;

  y1	=(ha*y1+sa[i0]*y2)*t;
  y2	=(ha*y2+sa[i0]*y3)*pt;
  y3	=(ha*y3+sa[i0]*y4)*pt*t;
  y4	*=ha*pt*pt;

  *df	=2.*(y0+y1+y2+y3+y4);

  r0	=y0+cr2*y1+cr3*y2+cr4*y3+cr5*y4;
  r1	=(cr2*y0+cr3*y1+cr4*y2+cr5*y3+cr6*y4);
  r2	=(cr3*y0+cr4*y1+cr5*y2+cr6*y3+cr7*y4);
  r3	=(cr4*y0+cr5*y1+cr6*y2+cr7*y3+cr8*y4);

  q0	=aT[i0];
  q1	=ha*daT[i0]*t;
  q3	=ha*daT[i1]*pt;

  pt	*=aT[i1]-q0;
  q2	=3.*pt-2.*q1*t-q3;
  q3	=(q3+q1*t-2.*pt)*t;

  *df	*=q0+q1+q2+q3;
  *f	=ha*t*(r0*q0+r1*q1+r2*q2+r3*q3);
  return(0);
}

int ZeroOfHermit(double *ts, double *df,int i)
{
  int i0,i1;
  double p0,p1,p2,p3;
  double a,b,c;
  double Q,R;

  i0	=i;
  i1	=i0+1;
  p0	=aT[i0];
  p1	=ha*daT[i0];
  p3	=ha*daT[i1];
  a	=aT[i1]-p0;
  p2	=3.*a-2.*p1-p3;
  p3	+=p1-2.*a;

  if(p3	== 0.){
    if(p2 == 0.){
      if(p1 == 0. || (R	=-p0/p1) < 0. || R > 1.){
	return(1);
      }
      *ts	=R;
      *df	=rha*p1;
      return(0);
    }
    a	=0.5*p1;
    c	=a*a-p2*p1;
    if(c < 0) return(1);
    c	=sqrt(c);
    R	=a > 0. ? -p1/(c+a) : (c-a)/p2;
    if(0. <= R && R <= 1.){
      *ts	=R;
      *df	=rha*(p1+2.*p2*R);
      return(0);
    }
    R	=a > 0. ?  -(c+a)/p2 : p1/(c-a);
    if(0. <= R && R <= 1.){
      *ts	=R;
      *df	=rha*(p1+2.*p2*R);
      return(0);
    }
    return(1);
  }
  else{
    R	=1./p3;
    a	=p2*cr3*R;
    b	=p1*R;
    c	=p0*R;
    
    Q	=a*a-cr3*b;
    R	=a*a*a+cr2*(c-a*b);
    b	=Q*Q*Q;
    if(R*R < b){
      b	=sqrt(Q);
      c	=cr3*acos(R/(Q*b));
      b	*=-2.;
      R	=b*cos(c)-a;
      if(0. <= R && R <= 1.){
	*ts	=R;
	*df	=rha*(p1+(2.*p2+3.*p3*R)*R);
	return(0);
      }
      R	=b*cos(c+cr3*c2gp)-a;
      if(0. <= R && R <= 1.){
	*ts	=R;
	*df	=rha*(p1+(2.*p2+3.*p3*R)*R);
	return(0);
      }
      R	=b*cos(c-cr3*c2gp)-a;
      if(0. <= R && R <= 1.){
	*ts	=R;
	*df	=rha*(p1+(2.*p2+3.*p3*R)*R);
	return(0);
      }
      return(1);
    }
    else{
      c	=R+sqrt(R*R-b);
      R	=c > 0. ? -pow(R+sqrt(R*R-b),cr3) : 0.;
      R	+=R != 0. ? Q/R-a : -a;
      if(0. <= R && R <= 1.){
	*ts	=R;
	*df	=rha*(p1+(2.*p2+3.*p3*R)*R);
	return(0);
      }
      return(1);
    }
  }
  return(0);
}
#endif/*stg_Hermit*/

/***************** ESI Initiation routines ****************************/
#ifndef stg_Initiation
int InitLocalConstants()
{
  cr2	=0.5;
  cr3	=1./3.;
  cr4	=0.25;
  cr5	=0.2;
  cr6	=1./6.;
  cr7	=1./7.;
  cr8	=0.125;
  cr12	=1./12.;
  cgp_4	=atan((double)1.);
  cgp	=4.*cgp_4;
  c2gp	=8.*cgp_4;
  cr2gp	=1./c2gp;
  cgm0	=0.4*cgp;
  crgm0	=1./cgm0;
  cbsp2ne=crgm0/1.6022e-2;

  kD	=sizeof(double);
  kI	=sizeof(int);
  return(0);
}

int SetESI(int i)
{
  ESI *lE;
  lE	=sESI+i;
  lE->Na	=Na;
  lE->Na1	=Na1;
  lE->Np	=Np;
  lE->Np1	=Np1;

  lE->gq	=gq;
  lE->sa	=sa;
  lE->gF	=gF;
  lE->dgF	=dgF;
  lE->d2gF	=d2gF;
  lE->gY	=gY;
  lE->dgY	=dgY;
  lE->d2gY	=d2gY;
  lE->aT	=aT;
  lE->daT	=daT;
  lE->aP	=aP;
  lE->daP	=daP;
  lE->aF	=aF;
  lE->daF	=daF;

  lE->bsp	=bsp;
  lE->dbsp	=dbsp;

  lE->sr	=sr;
  lE->sra	=sra;
  lE->srq	=srq;
  lE->sraq	=sraq;
  lE->sz	=sz;
  lE->sza	=sza;
  lE->szq	=szq;
  lE->szaq	=szaq;
  lE->aB	=aB;
  lE->aBa	=aBa;
  lE->aBq	=aBq;
  lE->aBaq	=aBaq;
  lE->gH	=gH;
  lE->gHa	=gHa;
  lE->gHq	=gHq;
  lE->gHaq	=gHaq;
  lE->gHqq	=gHqq;
  lE->gHaqq	=gHaqq;
  return(0);
}

int ESISetMem(double *gq0, int na1,int np1)
{
  int i;

  Np1	=np1;
  Na1	=na1;
  Na	=Na1-1;
  Np	=Np1-1;

  InitLocalConstants();
  gq	=gq0;
  sa	=gq	+Np1;

  gF	=sa	+Na1;
  dgF	=gF	+Na1;
  d2gF	=dgF	+Na1;

  gY	=d2gF	+Na1;
  dgY	=gY	+Na1;
  d2gY 	=dgY	+Na1;

  aT	=d2gY	+Na1;
  daT	=aT	+Na1;

  aP	=daT	+Na1;
  daP	=aP	+Na1;

  aF	=daP	+Na1;
  daF	=aF	+Na1;

  bsp	=daF	+Na1;
  dbsp	=bsp	+Na1;

  i	=Na1*Np1;
  sr	=dbsp	+Na1;
  sra	=sr	+i;
  srq	=sra	+i;
  sraq	=srq	+i;

  sz	=sraq	+i;
  sza	=sz	+i;
  szq	=sza	+i;
  szaq	=szq	+i;

  aB	=szaq	+i;
  aBa	=aB	+i;
  aBq	=aBa	+i;
  aBaq	=aBq	+i;

  gH	=aBaq	+i;
  gHa	=gH	+i;
  gHq	=gHa	+i;
  gHaq	=gHq	+i;
  gHqq	=gHaq	+i;
  gHaqq	=gHqq	+i;

  SetESI(0);
  return(0);
}

int ESIfree()
{
  if(gq != NULL) free(gq);
  return(0);
}

int ESIMemAlloc(int na1,int np1)
{
  if(gq != NULL) free(gq);

  Na1	=na1;
  Np1	=np1;
  Na	=Na1-1;
  Np	=Np1-1;
  Npa	=Np1*Na1;
  gq	=(REAL*)malloc((np1+(15+18*np1)*na1)*sizeof(REAL));
  if(gq == NULL){
    printf("Failure: No memory for ESI\n");
    return(1);
  }
  ESISetMem(gq,na1,np1);
  return(0);
}

int ESIInit()
{
  int i0,i1;
  int i,ii,j,ji,ji1;

  double y0,y1,y2,y3,y4;
  double q,q0,q1,q2,q3;
  double r0,r1,r2,r3;

  cgq0	=gq[Np];
  crgq0	=1./cgq0;
  hq	=gq[1]-gq[0];
  rhq	=1./hq;
  ha	=sa[1]-sa[0];
  rha	=1./ha;
  ji	=0;
  for(i=0; i < Na1; i++){
    gH[ji]	=0.;
    gHa[ji]	=0.;
    for(j=1; j < Np1; j++){
      ji1	=ji;
      ji++;
      gH[ji]	=gH[ji1]+((gHq[ji1]+gHq[ji])+cr6*hq*(gHqq[ji1]-gHqq[ji]))
	*hq*0.5;
      gHa[ji]	=gHa[ji1]+((gHaq[ji1]+gHaq[ji])
			   +cr6*hq*(gHaqq[ji1]-gHaqq[ji]))*hq*0.5;
    }
    ji++;
  }

  gF[0]	=0.;
  gY[0]	=0.;
  bsp[0]=0.;
  aF[0]=0.;
  for(i1=1,i0=0; i1 < Na1; i1++,i0++){
    y0	=ha*sa[i0];

    y1	=dgF[i0];
    y2	=ha*d2gF[i0];
    y4	=ha*d2gF[i1];
    q0	=dgF[i1]-dgF[i0];
    y3	=3.*q0-2.*y2-y4;
    y4	+=y2-2.*q0;

    q0	=sa[i0]*rha;
    y1	+=q0*y2;
    y2	+=q0*y3;
    y3	+=q0*y4;
    gF[i1]	=gF[i0]+y0*dgF[i0]+ha*ha*(cr2*y1+cr3*y2+cr4*y3+cr5*y4);

    y0	=sa[i0]*dgY[i0];
    y1	=dgY[i0];
    y2	=ha*d2gY[i0];
    y4	=ha*d2gY[i1];
    q0	=dgY[i1]-y1;
    y3	=3.*q0-2.*y2-y4;
    y4	+=y2-2.*q0;

    q0	=sa[i0]*rha;
    y1	+=q0*y2;
    y2	+=q0*y3;
    y3	+=q0*y4;
    gY[i1]	=gY[i0]+ha*(y0+ha*(cr2*y1+cr3*y2+cr4*y3+cr5*y4));

    r0	=y0+ha*(cr2*y1+cr3*y2+cr4*y3+cr5*y4);
    r1	=cr2*y0+ha*(cr3*y1+cr4*y2+cr5*y3+cr6*y4);
    r2	=cr3*y0+ha*(cr4*y1+cr5*y2+cr6*y3+cr7*y4);
    r3	=cr4*y0+ha*(cr5*y1+cr6*y2+cr7*y3+cr8*y4);

    q0	=aT[i0];
    q1	=ha*daT[i0];
    q3	=ha*daT[i1];
    q	=aT[i1]-q0;
    q2	=3.*q-2.*q1-q3;
    q3	+=q1-2.*q;
    aF[i1]	=aF[i0]+2.*ha*(r0*q0+r1*q1+r2*q2+r3*q3);

    q0	=aP[i0];
    q1	=ha*daP[i0];
    q3	=ha*daP[i1];
    q	=aP[i1]-q0;
    q2	=3.*q-2.*q1-q3;
    q3	+=q1-2.*q;
    bsp[i1]	=bsp[i0]   +ha*(r0*q0+r1*q1+r2*q2+r3*q3);
  }

  q1	=-bsp[Na];
  for(i=0; i < Na1; i++){
    bsp[i]	+=q1;
    dbsp[i]	=aP[i]*dgY[i]*sa[i];
  }
  aRBtor	=1.;
  if(RBtor >= 0.){
    r1	=RBtor*RBtor-aF[Na];
    for(i=0; i < Na1; i++){
      aF[i]	=sqrt(aF[i]+r1);
      daF[i]	=aT[i]*dgY[i]*sa[i]/aF[i];
    }
  }
  else{ /* RFP Case */
    i	=Na;
    while(i){
      i--;
      if(ZeroOfHermit(&aRBtor,&r1,i) == 0){
	EPPAHermit(&r0,&r1,aRBtor,i);
	aRBtor	=sa[i]+ha*aRBtor;
	r0	+=aF[i];
	break;
      }
    }
    for(i=0; i < Na1; i++){
      aF[i]	=sqrt(fabs(aF[i]-r0));
      if(sa[i] > aRBtor) aF[i]=-aF[i];
      if(sa[i] != aRBtor){
	daF[i]	=aT[i]*dgY[i]*sa[i]/aF[i];
      }
      else{
	daF[i]	=sqrt(fabs(daT[i]*dgY[i]*sa[i]));
      }
    }
    RBtor	=aF[Na];
  }
  return(0);
}

int Shmem2ESI(int *key)
{
  static int ShmID=0;
  static int *ShMemI=NULL;
  static char *lShm=NULL;
  double *ld;

  /* Init Shared memory */
  ShmID	=shmget((key_t)(*key),0, 0444 | IPC_CREAT);

  /* attach shared memory to process */
  lShm	=shmat(ShmID,NULL,0);
  ShMemI	=(int*)lShm;
  Na	=ShMemI[1];
  Np	=ShMemI[2];
  ld	=(double*)(ShMemI+2);
  RBtor	=*ld++;
  ESISetMem(ld,Na1,Np1);
  return(0);
}
#endif/*stg_Initiation*/

/***************** ESI Reading Routines *****************/
#ifndef stg_ReadingRoutines
static int nFA=0,kFA=0;
static FILE *lf,*lF;
char FNm[0x80],FNmB[0x80];

int FindGroup(char *Name, char *strR, char *strF,int nAttempt)
{
  char ch,*lc;
  int k,iA;
  
  iA	=0;
  while(iA < nAttempt){
    lc	=Name;
    while(kFA < nFA && *lc){
      ch	=(char)fgetc(lF);
      kFA++;
      if(ch == *lc){
	lc++;
      }
      else{
	lc	=Name;
      }
    }
    if(kFA < nFA){
      iA	=nAttempt;
    }
    else{
      rewind(lF);
      kFA	=0;
      iA++;
      if(iA == nAttempt){
	printf("Failure: %s: %s - record is absent\n",FNm,Name);
	return(1);
      }
    }
  }

  do{
    ch	=(char)fgetc(lF);
    kFA++;
  }while(kFA < nFA && isspace(ch));
  if(kFA == nFA || ch != '['){
    printf("Failure: %s: '[' - is absent after %s\n",FNm,Name);
    return(2);
  }
  lc	=strR;
  do{
    *lc=(char)fgetc(lF);
    kFA++;
    if(*lc == ']') break;
    else{
      if(!isspace(*lc)) lc++;
    }
  }while(kFA < nFA && *lc != ']');

  if(*lc != ']'){
    *lc	=0;
    *strF	=0;
    printf("Failure: %s: ']' - is absent after '['\n",FNm);
    return(3);
  };
  *lc	=0;
  do{
    ch	=(char)fgetc(lF);
    kFA++;
  }while(kFA < nFA && isspace(ch));

  if(kFA == nFA || ch != '('){
    printf("Failure: %s: '(' - is absent after [%s]\n",FNm,strR);
    return(4);
  }

  lc	=strF;
  while(kFA < nFA){
    *lc=(char)fgetc(lF);
    kFA++;
    if(*lc == ')') break;
    else{
      if(!isspace(*lc)) lc++;
    }
  }
  if(*lc != ')'){
    *lc	=0;
    printf("Failure: %s: ')' - is absent after [%s]('\n",FNm,strR);
    return(5);
  };
  *lc	='\0';
  ch	=0;
  while(kFA < nFA && ch != '\n'){
    ch=(char)fgetc(lF);
    kFA++;
  }
  return(0);
}

int ReadGroup(char *Name,char *strF, char **lv,int *dl,int n,int k)
{
  int i,j,m;
  char ln[256],*lc,*ls;
  char ch;

  for(i=0; i < n; i++){
    ls	=strF;
    j	=0;
    while(j < k){
      while(isspace(*ls)) ls++;
      m	=isdigit(*ls) ? 0 : 1;
      while(isdigit(*ls)){
	m	=m*10+(*ls)-'0';
	ls++;
      }
      if(*ls == '%'){
	ls++;
	while(m){
	  m--;
	  while(kFA < nFA){
	    ch	=(char)fgetc(lF);
	    kFA++;
	    if(ch == '!'){
	      while(kFA < nFA &&  ch != '\n'){
		ch	=(char)fgetc(lF);
		kFA++;
	      }
	    }
	    if(!isspace(ch)) break;
	  }
	  lc	=ln;
	  *lc++	=ch;
	  while(kFA < nFA){
	    ch	=(char)fgetc(lF);
	    kFA++;
	    if(isspace(ch)) break;
	    *lc++=ch;
	  }
	  *lc	='\0';
	  switch(*ls){
	  case 'd':
	    *(int*)lv[j]	=strtol(ln,&lc,10);
	    break;
	  case 'e':
	    *(double*)lv[j]	=strtod(ln,&lc);
	    break;
	  case 's':
	    strcpy((char*)lv[j],ln);
	    break;
	  }
	  if(*lc != '\0'){
	    printf("Failure: %s %s wrong [%d][%d]-th data\n",FNm,Name,i,j);
	    return(1);
	  }
	  lv[j]	+=dl[j];
	}
	j++;
	ls++;
      }
      else{
	while(kFA < nFA){
	  ch	=(char)fgetc(lF);
	  kFA++;
	  if(ch == '!'){
	    while(kFA < nFA &&  ch != '\n'){
	      ch	=(char)fgetc(lF);
	      kFA++;
	    }
	  }
	  if(!isspace(ch)) break;
	}
	while(*ls != '\0' && ch == *ls && *ls != '%' && !isspace(*ls)){
	  ls++;
	  if(kFA < nFA){
	    ch=(char)fgetc(lF);
	    kFA++;
	  }
	}
      }      
    }
    while(kFA < nFA && ch != '\n'){
      ch=(char)fgetc(lF);
      kFA++;
    }
  }
  return(0);
}

int Group0()
{
  int iG=0;
  int i,j,n,k;
  char strR[64],strF[64];
  char *lc;
  char *lv[8];
  int dl[8]={0,0,0,0,0,0};
  char Nm[0x20];
  
  if(FindGroup(GrName[iG],strR,strF,2)){
    printf("Failure: %s.%s - wrong head format\n",FNm,GrName[iG]);
    return(1);
  }
  lc	=strF;
  k	=0;
  while(*lc) if(*lc++ == '%') k++;
  if(sscanf(strR,"%d",&n) < 1){
    printf("Failure: %s.%s%s - no number\n",FNm,GrName[iG],strR);
    return(1);
  }
  lv[0]	=(char*)&j;
  lv[1]	=(char*)&i;
  lv[2]	=(char*)&ID;
  lv[3]	=(char*)&Rext;
  lv[4]	=(char*)&RBtor;
  lv[5]	=(char*)&Nm;
  k	=6;
  if(ReadGroup(GrName[iG],strF,lv,dl,n,k)) return(1);
  
  if(gq == NULL || Np1 != j || Na1 != i){
    if(gq != NULL) ESIfree();
    if(ESIMemAlloc(i,j)) return(1);
  } 
  return(0);
}

int Group1()
{
  int iG=1;
  int j,n,k,m;
  char strR[64],strF[64];
  char *lc;
  char *lm;
  char *lv[8];
  int dl[8]={0,0,0,0,0,0};
  double dq;
  
  if(FindGroup(GrName[iG],strR,strF,1)){
    rewind(lF);
    dq	=c2gp/Np;
    for(j=0; j < Np1; j++) gq[j]=dq*j;
    return(0);
  }
  lc	=strF;
  k	=0;
  while(*lc) if(*lc++ == '%') k++;
  if(sscanf(strR,"%d",&n) < 1){
    printf("Failure: %s.%s%s - number is absent\n",FNm,GrName[iG],strR);
    return(1);
  }

  lc	=strF;
  m	=isdigit(*lc) ? 0 : 1;
  while(isdigit(*lc)) m=m*10+(*lc++)-'0';
  if(n*m != Np1 && n*m > 2){
    printf("Failure: %s.%s[%s](%s) - %d * %d != %d\n"
	   ,FNm,GrName[iG],strR,strF,n,m,Np1);
    return(1);
  }
  m	*=n;

  if(m){
    lv[0]	=(char*)gq;
    dl[0]	=kD;
    if(ReadGroup(GrName[iG],strF,lv,dl,n,k)) return(1);
  }
  if(m == Np1) return(0);

  if(m < 2){
    dq	=m ? gq[0]/Np : c2gp/Np;
    gq[0]	=0.;
  }
  else{
    if(m == 2){
      dq	=(gq[1]-gq[0])/Np;
    }
  }

  if(dq == 0.){
    printf("Warning: %s.%s %10.3e == %10.3e - wrong interval\n"
	   ,FNm,GrName[iG],gq[1],gq[0]);
    gq[0]	=0.;
    dq	=c2gp/Np;
  }
  for(j=0; j < Np1; j++) gq[j]=gq[0]+dq*j;
  return(0);
}

int Group2()
{
  int iG=2;
  int i,n,k,m;
  char strR[64],strF[64];
  char *lc;
  char *lv[8];
  int dl[8]={0,0,0,0,0,0};
  double da;
  
  if(FindGroup(GrName[iG],strR,strF,1)){
    rewind(lF);
    da	=1./Na;
    for(i=0; i < Np1; i++) sa[i]=da*i;
    return(0);
  }
  lc	=strF;
  k	=0;
  while(*lc) if(*lc++ == '%') k++;
  if(sscanf(strR,"%d",&n) < 1){
    printf("Failure: %s.%s%s - number is absent\n",FNm,GrName[iG],strR);
    return(1);
  }

  lc	=strF;
  m	=isdigit(*lc) ? 0 : 1;
  while(isdigit(*lc)) m=m*10+(*lc++)-'0';
  if(n*m != Na1 && n*m > 2){
    printf("Failure: %s.%s[%s](%s) - %d * %d != %d\n"
	   ,FNm,GrName[iG],strR,strF,n,m,Na1);
    return(1);
  }
  m	*=n;

  if(m){
    lv[0]	=(char*)sa;
    dl[0]	=kD;
    if(ReadGroup(GrName[iG],strF,lv,dl,n,k)) return(1);
  }
  if(m == Na1) return(0);

  if(m < 2){
    da	=m ? sa[0]/Na : 1./Na;
    sa[0]	=0.;
  }
  else{
    if(m == 2){
      da	=(sa[1]-sa[0])/Na;
    }
  }

  if(da == 0.){
    printf("Warning: %s.%s %10.3e == %10.3e - wrong interval\n"
	   ,FNm,GrName[iG],sa[1],sa[0]);
    sa[0]	=0.;
    da	=1./Na;
  }
  for(i=0; i < Na1; i++) sa[i]=sa[0]+da*i;
  return(0);
}

int Group4()
{
  int iG=4;
  int i,n,k,m;
  char strR[64],strF[64];
  char *lc;
  char *lm;
  char *lv[8];
  int dl[8]={0,0,0,0,0,0};
  double da;
  
  if(FindGroup(GrName[iG],strR,strF,2)){
    printf("Failure: %s.%s - wrong head format\n",FNm,GrName[iG]);
    return(1);
  }
  lc	=strF;
  k	=0;
  while(*lc) if(*lc++ == '%') k++;
  if(sscanf(strR,"%d",&n) < 1){
    printf("Failure: %s.%s%s - number is absent\n",FNm,GrName[iG],strR);
    return(1);
  }

  lc	=strF;
  m	=isdigit(*lc) ? 0 : 1;
  while(isdigit(*lc)) m=m*10+(*lc++)-'0';
  if(n*m != Na1){
    printf("Failure: %s.%s[%s](%s) - %d * %d != %d\n"
	   ,FNm,GrName[iG],strR,strF,n,m,Na1);
    return(1);
  }
  lv[0]	=(char*)dgF;
  dl[0]	=kD;
  lv[1]	=(char*)d2gF;
  dl[1]	=kD;
  lv[2]	=(char*)dgY;
  dl[2]	=kD;
  lv[3]	=(char*)d2gY;
  dl[3]	=kD;
  if(ReadGroup(GrName[iG],strF,lv,dl,n,k)) return(1);
  return(0);
}

int Group5()
{
  int iG=5;
  int i,n,k,m;
  char strR[64],strF[64];
  char *lc;
  char *lv[8];
  int dl[8]={0,0,0,0,0,0};
  double da;
  
  if(FindGroup(GrName[iG],strR,strF,2)){
    printf("Failure: %s.%s - wrong head format\n",FNm,GrName[iG]);
    return(1);
  }
  lc	=strF;
  k	=0;
  while(*lc) if(*lc++ == '%') k++;
  if(sscanf(strR,"%d",&n) < 1){
    printf("Failure: %s.%s%s - number is absent\n",FNm,GrName[iG],strR);
    return(1);
  }

  lc	=strF;
  m	=isdigit(*lc) ? 0 : 1;
  while(isdigit(*lc)) m=m*10+(*lc++)-'0';
  if(n*m != Na1){
    printf("Failure: %s.%s[%s](%s) - %d * %d != %d\n"
	   ,FNm,GrName[iG],strR,strF,n,m,Na1);
    return(1);
  }
  lv[0]	=(char*)aT;
  dl[0]	=kD;
  lv[1]	=(char*)daT;
  dl[1]	=kD;
  lv[2]	=(char*)aP;
  dl[2]	=kD;
  lv[3]	=(char*)daP;
  dl[3]	=kD;
  if(ReadGroup(GrName[iG],strF,lv,dl,n,k)) return(1);
  return(0);
}

int Group6()
{
  int iG=6;
  int i,n,k,m,L;
  long int lA;
  char strR[64],strF[64];
  char *lc;
  char *lv[8];
  int dl[8]={0,0,0,0,0,0};
  
  if(FindGroup(GrName[iG],strR,strF,2)){
    printf("Failure: %s.%s - wrong head format\n",FNm,GrName[iG]);
    return(1);
  }
  lc	=strF;
  k	=0;
  while(*lc) if(*lc++ == '%') k++;
  lA	=0;
  *FNmB	=0;
  m	=sscanf(strR,"%d @ %ld : %s",&n,&lA,FNmB);
  if(m == 0){
    printf("Failure: %s.%s%s - number is absent\n",FNm,GrName[iG],strR);
    return(1);
  }

  lc	=strF;
  L	=isdigit(*lc) ? 0 : 1;
  while(isdigit(*lc)) L=L*10+(*lc++)-'0';
  if(n*L != Npa){
    printf("Failure: %s.%s[%s](%s) - %d*%d != %d\n"
	   ,FNm,GrName[iG],strR,strF,n,L,Npa);
    return(1);
  }

  lv[0]	=(char*)sr;
  lv[1]	=(char*)sra;
  lv[2]	=(char*)srq;
  lv[3]	=(char*)sraq;
  switch(m){
  case 1:
    for(i=0; i < 4; i++) dl[i]=kD;
    if(ReadGroup(GrName[iG],strF,lv,dl,n,k)) return(1);
    break;
  case 2:
    if(lf == NULL){
      printf("Failure: %s.%s%s - no binary file name\n",FNm,GrName[iG],strR);
      return(1);
    }
    fseek(lf,lA,SEEK_SET);
    for(i=0; i < 4; i++) dl[i]=kD*Npa;
    while(n){
      n--;
      for(i=0; i < 4; i++){
	fread(lv[i],kD,Np1,lf);
	lv[i]	+=dl[i];
      }
    }
    break;
  case 3:
    if(lf == NULL){
      struct stat statbuf;
      if(stat(FNmB,&statbuf)){
	printf("Failure: %s.%s%s - no binary file\n",FNm,GrName[iG],strR);
	return(1);
      }
    }
    lf	=fopen(FNmB,"r");
    fseek(lf,lA,SEEK_SET);
    for(i=0; i < 4; i++) dl[i]=kD*Npa;
    while(n){
      n--;
      for(i=0; i < 4; i++){
	fread(lv[i],kD,Np1,lf);
	lv[i]	+=dl[i];
      }
    }
    break;
  default:
    break;
  }
  return(0);
}

int Group7()
{
  int iG=7;
  int i,n,k,m,L;
  long int lA;
  char strR[64],strF[64];
  char *lc;
  char *lv[8];
  int dl[8]={0,0,0,0,0,0};
  
  if(FindGroup(GrName[iG],strR,strF,2)){
    printf("Failure: %s.%s - wrong head format\n",FNm,GrName[iG]);
    return(1);
  }
  lc	=strF;
  k	=0;
  while(*lc) if(*lc++ == '%') k++;
  lA	=0;
  *FNmB	=0;
  m	=sscanf(strR,"%d @ %ld : %s",&n,&lA,FNmB);
  if(m == 0){
    printf("Failure: %s.%s%s - number is absent\n",FNm,GrName[iG],strR);
    return(1);
  }

  lc	=strF;
  L	=isdigit(*lc) ? 0 : 1;
  while(isdigit(*lc)) L=L*10+(*lc++)-'0';
  if(n*L != Npa){
    printf("Failure: %s.%s[%s](%s) - %d*%d != %d\n"
	   ,FNm,GrName[iG],strR,strF,n,L,Npa);
    return(1);
  }

  lv[0]	=(char*)sz;
  lv[1]	=(char*)sza;
  lv[2]	=(char*)szq;
  lv[3]	=(char*)szaq;
  switch(m){
  case 1:
    for(i=0; i < 4; i++) dl[i]=kD;
    if(ReadGroup(GrName[iG],strF,lv,dl,n,k)) return(1);
    break;
  case 2:
    if(lf == NULL){
      printf("Failure: %s.%s%s - no binary file name\n",FNm,GrName[iG],strR);
      return(1);
    }
    fseek(lf,lA,SEEK_SET);
    for(i=0; i < 4; i++) dl[i]=kD*Npa;
    while(n){
      n--;
      for(i=0; i < 4; i++){
	fread(lv[i],kD,Np1,lf);
	lv[i]	+=dl[i];
      }
    }
    break;
  case 3:
    if(lf == NULL){
      struct stat statbuf;
      if(stat(FNmB,&statbuf)){
	printf("Failure: %s.%s%s - no binary file\n",FNm,GrName[iG],strR);
	return(1);
      }
    }
    lf	=fopen(FNmB,"r");
    fseek(lf,lA,SEEK_SET);
    for(i=0; i < 4; i++) dl[i]=kD*Npa;
    while(n){
      n--;
      for(i=0; i < 4; i++){
	fread(lv[i],kD,Np1,lf);
	lv[i]	+=dl[i];
      }
    }
    break;
  default:
    break;
  }
  return(0);
}

int Group8()
{
  int iG=8;
  int i,n,k,m,L;
  long int lA;
  char strR[64],strF[64];
  char *lc;
  char *lv[8];
  int dl[8]={0,0,0,0,0,0};
  
  if(FindGroup(GrName[iG],strR,strF,2)){
    printf("Failure: %s.%s - wrong head format\n",FNm,GrName[iG]);
    return(1);
  }
  lc	=strF;
  k	=0;
  while(*lc) if(*lc++ == '%') k++;
  lA	=0;
  *FNmB	=0;
  m	=sscanf(strR,"%d @ %ld : %s",&n,&lA,FNmB);
  if(m == 0){
    printf("Failure: %s.%s%s - number is absent\n",FNm,GrName[iG],strR);
    return(1);
  }

  lc	=strF;
  L	=isdigit(*lc) ? 0 : 1;
  while(isdigit(*lc)) L=L*10+(*lc++)-'0';
  if(n*L != Npa){
    printf("Failure: %s.%s[%s](%s) - %d*%d != %d\n"
	   ,FNm,GrName[iG],strR,strF,n,L,Npa);
    return(1);
  }

  lv[0]	=(char*)aB;
  lv[1]	=(char*)aBa;
  lv[2]	=(char*)aBq;
  lv[3]	=(char*)aBaq;
  switch(m){
  case 1:
    for(i=0; i < 4; i++) dl[i]=kD;
    if(ReadGroup(GrName[iG],strF,lv,dl,n,k)) return(1);
    break;
  case 2:
    if(lf == NULL){
      printf("Failure: %s.%s%s - no binary file name\n",FNm,GrName[iG],strR);
      return(1);
    }
    fseek(lf,lA,SEEK_SET);
    for(i=0; i < 4; i++) dl[i]=kD*Npa;
    while(n){
      n--;
      for(i=0; i < 4; i++){
	fread(lv[i],kD,Np1,lf);
	lv[i]	+=dl[i];
      }
    }
    break;
  case 3:
    if(lf == NULL){
      struct stat statbuf;
      if(stat(FNmB,&statbuf)){
	printf("Failure: %s.%s%s - no binary file\n",FNm,GrName[iG],strR);
	return(1);
      }
    }
    lf	=fopen(FNmB,"r");
    fseek(lf,lA,SEEK_SET);
    for(i=0; i < 4; i++) dl[i]=kD*Npa;
    while(n){
      n--;
      for(i=0; i < 4; i++){
	fread(lv[i],kD,Np1,lf);
	lv[i]	+=dl[i];
      }
    }
    break;
  default:
    break;
  }
  return(0);
}

int Group9()
{
  int iG=9;
  int i,n,k,m,L;
  long int lA;
  char strR[64],strF[64];
  char *lc;
  char *lv[8];
  int dl[8]={0,0,0,0,0,0};
  
  if(FindGroup(GrName[iG],strR,strF,2)){
    printf("Failure: %s.%s - wrong head format\n",FNm,GrName[iG]);
    return(1);
  }
  lc	=strF;
  k	=0;
  while(*lc) if(*lc++ == '%') k++;
  lA	=0;
  *FNmB	=0;
  m	=sscanf(strR,"%d @ %ld : %s",&n,&lA,FNmB);
  if(m == 0){
    printf("Failure: %s.%s%s - number is absent\n",FNm,GrName[iG],strR);
    return(1);
  }

  lc	=strF;
  L	=isdigit(*lc) ? 0 : 1;
  while(isdigit(*lc)) L=L*10+(*lc++)-'0';
  if(n*L != Npa){
    printf("Failure: %s.%s[%s](%s) - %d*%d != %d\n"
	   ,FNm,GrName[iG],strR,strF,n,L,Npa);
    return(1);
  }

  lv[0]	=(char*)gHq;
  lv[1]	=(char*)gHaq;
  lv[2]	=(char*)gHqq;
  lv[3]	=(char*)gHaqq;
  switch(m){
  case 1:
    for(i=0; i < 4; i++) dl[i]=kD;
    if(ReadGroup(GrName[iG],strF,lv,dl,n,k)) return(1);
    break;
  case 2:
    if(lf == NULL){
      printf("Failure: %s.%s%s - no binary file name\n",FNm,GrName[iG],strR);
      return(1);
    }
    fseek(lf,lA,SEEK_SET);
    for(i=0; i < 4; i++) dl[i]=kD*Npa;
    while(n){
      n--;
      for(i=0; i < 4; i++){
	fread(lv[i],kD,Np1,lf);
	lv[i]	+=dl[i];
      }
    }
    break;
  case 3:
    if(lf == NULL){
      struct stat statbuf;
      if(stat(FNmB,&statbuf)){
	printf("Failure: %s.%s%s - no binary file\n",FNm,GrName[iG],strR);
	return(1);
      }
    }
    lf	=fopen(FNmB,"r");
    fseek(lf,lA,SEEK_SET);
    for(i=0; i < 4; i++) dl[i]=kD*Npa;
    while(n){
      n--;
      for(i=0; i < 4; i++){
	fread(lv[i],kD,Np1,lf);
	lv[i]	+=dl[i];
      }
    }
    break;
  default:
    break;
  }
  return(0);
}

int File2ESI(char* FName)
{
  struct stat statbuf;
  
  lf	=NULL;
  strcpy(FNm,FName);
  if(stat(FNm,&statbuf)){
    printf("Failure: <%s> is absent !!!\n",FNm);
    return(1);
  }
  nFA	=statbuf.st_size;
  kFA	=0;
  lF	=fopen(FNm,"r");

  if(Group0() || Group1() || Group2() || Group4() || Group5()
     || Group6() || Group7() || Group8() || Group9()){
    fclose(lF);
    return(1);
  }  
  if(lf != NULL) fclose(lf);
  fclose(lF);

  ESIInit();
  return(0);
}

int file2esc_(char* FName)
{
  struct stat statbuf;
  char *lc,*ls;

  ls	=FName;
  while(ls-FName < 0x80 && isspace(*ls)) ls++;
  lc	=FNm;
  while(ls-FName && lc-FNm < 0x80 && !isspace(*ls)) *lc++ =*ls++;
  *lc	='\0';

  lf	=NULL;
  strcpy(FNm,FName);
  if(stat(FNm,&statbuf)){
    printf("Failure: <%s> is absent !!!\n",FNm);
    return(1);
  }
  nFA	=statbuf.st_size;
  kFA	=0;
  lF	=fopen(FNm,"r");

  if(Group0() || Group1() || Group2() || Group4() || Group5()
     || Group6() || Group7() || Group8() || Group9()){
    fclose(lF);
    return(1);
  }  
  if(lf != NULL) fclose(lf);
  fclose(lF);

  ESIInit();
  return(0);
}
#endif /*stg_ReadingRoutines*/

/***************** ESI link with user's FORTRAN-C space *****************/

#ifndef stg_Basic4User
int Link2ESI(REAL *XaF,REAL *XaFa
		,REAL *XgFa,REAL *XgFaa
		,REAL *XgYa,REAL *XgYaa
		,REAL *XT,REAL *XTa
		,REAL *XP,REAL *XPa
		,REAL *Xr,REAL *Xra,REAL *Xrq
		,REAL *Xz,REAL *Xza,REAL *Xzq
		,REAL *XB,REAL *XBa,REAL *XBq
		,REAL *Xgh,REAL *Xgha,REAL *Xghq
		,int *Xsw)
{
  gFa	=XgFa;
  gFaa	=XgFaa;
  gYa	=XgYa;
  gYaa	=XgYaa;

  T	=XT;
  Ta	=XTa;
  P	=XP;
  Pa	=XPa;

  F	=XaF;
  Fa	=XaFa;

  r	=Xr;
  ra	=Xra;
  rq	=Xrq;
  z	=Xz;
  za	=Xza;
  zq	=Xzq;
  B	=XB;
  Ba	=XBa;
  Bq	=XBq;
  gh	=Xgh;
  gha	=Xgha;
  ghq	=Xghq;
  sw	=Xsw;
  return(0);
}

/***************** Main Reconstruction Routines **********************/
int esi2gFgY(double *gf, double *gy,double *ap,int n)
{
  int i,i0,i1;
  double t,a;
  double q,q0,q1,q2,q3,q4,A0,A1,A2,A3;

  for(i=0; i < n; i++){
    a	=ap[i];
    if(a < 0.)   a=-a;
    i0	=(a-sa[0])*rha;
    if(i0 == 0 && a == 0. || i0 == Na){
      gf[i]	=gF[i0];
      gy[i]	=gY[i0];
    }
    else{
      i1	=i0+1;
      t	=(a-sa[i0])*rha;

      q0	=rha*sa[i0]*t;
      q1	=t*t;
      A0	=q0+cr2*q1;
      A1	=(cr2*q0+cr3*q1)*t;
      A2	=(cr3*q0+cr4*q1)*q1;
      A3	=(cr4*q0+cr5*q1)*q1*t;
      q0	=A0-3.*A2+2.*A3;
      q1	=3.*A2-2.*A3;
      q2	=A1-2.*A2+A3;
      q3	=-A2+A3;
      gf[i]	=gF[i0]
	+ha*ha*(q0*dgF[i0]+q1*dgF[i1]+q2*ha*d2gF[i0]+q3*ha*d2gF[i1]);
      gy[i]	=gy[i0]
	+ha*ha*(q0*dgY[i0]+q1*dgY[i1]+q2*ha*d2gY[i0]+q3*ha*d2gY[i1]);
    }
    gf[i]	*=c2gp;
    gy[i]	*=c2gp;
  }
  return(0);
}

int ESI2bsp(double *bsp0, double *dbsp0,double *ap,int n)
{
  int i,i0,i1;
  double a,dy;
  double t,pt;
  double y,y0,y1,y2,y3,y4;
  double q,q0,q1,q2,q3;
  double r0,r1,r2,r3;

  for(i=0; i < n; i++){
    a	=ap[i];
    if(a < 0.)   a=-a;
    i0	=(a-sa[0])*rha;
    if(i0 == 0 && a == 0. || i0 == Na){
      if(dbsp0 != NULL) dbsp0[i]=aP[i0]*dgY[i0]*sa[i0];
      bsp0[i]	=bsp[i0];
    }
    else{
      i1	=i0+1;
      t	=(a-sa[i0])*rha;
      pt	=t*t;
      a		=sa[i0];
      y0	=a*dgY[i0];
      y1	=dgY[i0];
      y2	=ha*d2gY[i0];
      y4	=ha*d2gY[i1];
      q0	=dgY[i1]-y1;
      y3	=3.*q0-2.*y2-y4;
      y4	+=y2-2.*q0;

      dy	=y1+(y2+(y3+y4*t)*t)*t;

      y1	=(ha*y1+a*y2)*t;
      y2	=(ha*y2+a*y3)*pt;
      y3	=(ha*y3+a*y4)*pt*t;
      y4	*=ha*pt*pt;

      r0	=y0+cr2*y1+cr3*y2+cr4*y3+cr5*y4;
      r1	=cr2*y0+cr3*y1+cr4*y2+cr5*y3+cr6*y4;
      r2	=cr3*y0+cr4*y1+cr5*y2+cr6*y3+cr7*y4;
      r3	=cr4*y0+cr5*y1+cr6*y2+cr7*y3+cr8*y4;

      pt	=t*t;
      q0	=aP[i0];
      q1	=ha*daP[i0]*t;
      q3	=ha*daP[i1]*pt;
      pt	*=aP[i1]-q0;
      q2	=3.*pt-2.*q1*t-q3;
      q3	=(q3+q1*t-2.*pt)*t;
      if(dbsp0 != NULL) dbsp0[i]=dy*(q0+q1+q2+q3);
      bsp0[i]	=bsp[i0]+ha*t*(r0*q0+r1*q1+r2*q2+r3*q3);
    }
  }
  return(0);
}

int ESI2Profiles(REAL *sF,REAL *sFa,	 
		       /* \baF,\R{d\baF}{da}*/
		       REAL *sgFa,REAL *sgFaa,
		       /* \R{d\bgF}{ada},(\R{d\bgF}{ada})'_a */
		       REAL *sgYa,REAL *sgYaa,
		       /* \R{d\bgY}{ada},(\R{d\bgY}{ada})'_a */
		       REAL *sT,REAL *sTa,
		       /* T=\baF\R{d\baF}{d\bgY},\R{dT}{da} */
		       REAL *sP,REAL *sPa
		       /* P=\R{d\bsp}{d\bgY},\R{dP}{da} */
		       ,REAL *a, int n)
{
  int k;
  double t,pt;
  double q0,q1,q2,q3;
  double d0,d1,d2,d3;

  for(k=0; k < n; k++){
    if(a[k] == 0){
      sF[k]	=aF[0];
      sFa[k]	=0.;
      sgFa[k]	=0.;
      sgFaa[k]	=dgF[0];
      sgYa[k]	=0.;
      sgYaa[k]	=dgY[0];
      sT[k]	=aT[0];
      sTa[k]	=0.;
      sP[k]	=aP[0];
      sPa[k]	=0.;
      sp[k]	=bsp[0];
      spa[k]	=0.;
    }
    else{
      A	=a[k];
      if(A < 0.) A=-A;
      i0	=(A-sa[0])*rha;
      if(i0 >= Na){
	i0=Na-1;
	A	=sa[Na];
      }
      i1	=i0+1;

      t		=(A-sa[i0])*rha;
      pt	=t*t;
      d1	=2.-2.*t;
      q1	=(1.+d1)*pt;
      q0	=1.-q1;
      d0	=t-pt;
      q3	=-d0*t*ha;
      q2	=q3+d0*ha;
      d0	*=6.*rha;
      d2	=pt-d1*t;
      d1	+=d2-1.;

      sgFa[k]	=q0*dgF[i0]+q1*dgF[i1]+q2*d2gF[i0]+q3*d2gF[i1];
      sgFaa[k]=d0*(dgF[i1]-dgF[i0])+d1*d2gF[i0]+d2*d2gF[i1];
      sgFaa[k]=A*(*sgFaa)+(*sgFa);
      sgYa[k]	=q0*dgY[i0]+q1*dgY[i1]+q2*d2gY[i0]+q3*d2gY[i1];
      sgYaa[k]=d0*(dgY[i1]-dgY[i0])+d1*d2gY[i0]+d2*d2gY[i1];
      sgYaa[k]=A*(*sgYaa)+(*sgYa);
      if(A != 0.){
	sgFa[k]	*=A;
	sgYa[k]	*=A;
      }
      sT[k]	=q0*aT[i0]+q1*aT[i1]+q2*daT[i0]+q3*daT[i1];
      sTa[k]	=d0*(aT[i1]-aT[i0])+d1*daT[i0]+d2*daT[i1];
      sP[k]	=q0*aP[i0]+q1*aP[i1]+q2*daP[i0]+q3*daP[i1];
      sPa[k]	=d0*(aP[i1]-aP[i0])+d1*daP[i0]+d2*daP[i1];
      sF[k]	=q0*aF[i0]+q1*aF[i1]+q2*daF[i0]+q3*daF[i1];
      sFa[k]	=d0*(aF[i1]-aF[i0])+d1*daF[i0]+d2*daF[i1];
      if(sp != NULL){
	sp[k]	=q0*bsp[i0]+q1*bsp[i1]+q2*dbsp[i0]+q3*dbsp[i1];
	spa[k]=d0*(bsp[i1]-bsp[i0])+d1*dbsp[i0]+d2*dbsp[i1];
      }
    }
  }
  return(0);
}

int ESI2All(REAL *a0,REAL *gq0,int n)
{
  /* Exception: at a=0 the routine returns:
     dr_q/a, dz_q/a, dB_q/a, dgF/a, dgY/a
     rather than
     dr_q=0, dz_q=0, dB_q=0, dgF=0, dgY=0 */
  int j0,j1,k;
  REAL a;
  double t,pt;
  double X0,X1,X2,X3;
  double x0,x1,x2;
  double Y0,Y1,Y2,Y3;
  double y0,y1,y2;
  double E0,E1,E2,E3;

  REAL f0,f1,fq0,fq1;
  for(k =0; k < n; k++){
    t	=gq0[k];
    j1	=t*crgq0;
    if(t < 0.) j1--;
    t	-=cgq0*j1;
    j0	=t*rhq;
    if(j0 >= Np){
      j0	-=Np;
      t	-=cgq0;
    }
    if(j0 < 0){
      j0	+=Np;
      t	+=cgq0;
    }
    j1	=j0+1;
    t	=(t-gq[j0])*rhq;

    pt	=t*t;

    Y3	=cr4*pt*pt;
    Y2	=t*pt;
    E0	=t-Y2+2.*Y3;
    E1	=Y2-2.*Y3;
    E2	=cr2*pt-2.*cr3*Y2+Y3;
    E3	=-cr3*Y2+Y3;

    y1	=2.-2.*t;
    Y1	=(1.+y1)*pt;
    Y0	=1.-Y1;
    y0	=t-pt;
    Y3	=-y0*t*hq;
    Y2	=Y3+y0*hq;
    y0	*=6.*rhq;
    y2	=pt-y1*t;
    y1	+=y2-1.;

    a	=a0[k];
    if(a < 0.) a=-a;
    i0	=(a-sa[0])*rha;
    if(a != 0.){
      if(i0 >= Na) i0=Na-1;
      i1	=i0+1;
      t	=(a-sa[i0])*rha;

      pt	=t*t;
      x1	=2.-2.*t;
      X1	=(1.+x1)*pt;
      X0	=1.-X1;
      x0	=t-pt;
      X3	=-x0*t*ha;
      X2	=X3+x0*ha;
      x0	*=6.*rha;
      x2	=pt-x1*t;
      x1	+=x2-1.;
      
      j00	=Np1*i0+j0;
      j01	=j00+1;
      j10	=j00+Np1;
      j11	=j10+1;
      
      /* r,r'_a,r'_q */
      f0	=X0*sr [j00]+X1*sr [j10]+X2*sra [j00]+X3*sra [j10];
      f1	=X0*sr [j01]+X1*sr [j11]+X2*sra [j01]+X3*sra [j11];
      fq0	=X0*srq[j00]+X1*srq[j10]+X2*sraq[j00]+X3*sraq[j10];
      fq1	=X0*srq[j01]+X1*srq[j11]+X2*sraq[j01]+X3*sraq[j11];
      r[k]	=Y0*f0+Y1*f1+Y2*fq0+Y3*fq1;
      rq[k]	=y0*(f1-f0)+y1*fq0+y2*fq1;
      f0	=x0*(sr [j10]-sr [j00])+x1*sra [j00]+x2*sra [j10];
      f1	=x0*(sr [j11]-sr [j01])+x1*sra [j01]+x2*sra [j11];
      fq0	=x0*(srq[j10]-srq[j00])+x1*sraq[j00]+x2*sraq[j10];
      fq1	=x0*(srq[j11]-srq[j01])+x1*sraq[j01]+x2*sraq[j11];
      ra[k]	=Y0*f0+Y1*f1+Y2*fq0+Y3*fq1;
      
      /* z,z'_a,z'_q */
      f0	=X0*sz [j00]+X1*sz [j10]+X2*sza [j00]+X3*sza [j10];
      f1	=X0*sz [j01]+X1*sz [j11]+X2*sza [j01]+X3*sza [j11];
      fq0	=X0*szq[j00]+X1*szq[j10]+X2*szaq[j00]+X3*szaq[j10];
      fq1	=X0*szq[j01]+X1*szq[j11]+X2*szaq[j01]+X3*szaq[j11];
      z[k]	=Y0*f0+Y1*f1+Y2*fq0+Y3*fq1;
      zq[k]	=y0*(f1-f0)+y1*fq0+y2*fq1;
      f0	=x0*(sz [j10]-sz [j00])+x1*sza [j00]+x2*sza [j10];
      f1	=x0*(sz [j11]-sz [j01])+x1*sza [j01]+x2*sza [j11];
      fq0	=x0*(szq[j10]-szq[j00])+x1*szaq[j00]+x2*szaq[j10];
      fq1	=x0*(szq[j11]-szq[j01])+x1*szaq[j01]+x2*szaq[j11];
      za[k]	=Y0*f0+Y1*f1+Y2*fq0+Y3*fq1;

      /* B,B'_a,B'_q */
      f0	=X0*aB [j00]+X1*aB [j10]+X2*aBa [j00]+X3*aBa [j10];
      f1	=X0*aB [j01]+X1*aB [j11]+X2*aBa [j01]+X3*aBa [j11];
      fq0	=X0*aBq[j00]+X1*aBq[j10]+X2*aBaq[j00]+X3*aBaq[j10];
      fq1	=X0*aBq[j01]+X1*aBq[j11]+X2*aBaq[j01]+X3*aBaq[j11];
      B[k]	=Y0*f0+Y1*f1+Y2*fq0+Y3*fq1;
      Bq[k]	=y0*(f1-f0)+y1*fq0+y2*fq1;
      f0	=x0*(aB [j10]-aB [j00])+x1*aBa [j00]+x2*aBa [j10];
      f1	=x0*(aB [j11]-aB [j01])+x1*aBa [j01]+x2*aBa [j11];
      fq0	=x0*(aBq[j10]-aBq[j00])+x1*aBaq[j00]+x2*aBaq[j10];
      fq1	=x0*(aBq[j11]-aBq[j01])+x1*aBaq[j01]+x2*aBaq[j11];
      Ba[k]	=Y0*f0+Y1*f1+Y2*fq0+Y3*fq1;

      /* gh,gh'_a,gh'_q */
      f0	=X0*gHq [j00]+X1*gHq [j10]+X2*gHaq [j00]+X3*gHaq [j10];
      f1	=X0*gHq [j01]+X1*gHq [j11]+X2*gHaq [j01]+X3*gHaq [j11];
      fq0	=X0*gHqq[j00]+X1*gHqq[j10]+X2*gHaqq[j00]+X3*gHaqq[j10];
      fq1	=X0*gHqq[j01]+X1*gHqq[j11]+X2*gHaqq[j01]+X3*gHaqq[j11];
      ghq[k]	=Y0*f0+Y1*f1+Y2*fq0+Y3*fq1;
      gh[k]	=X0*gH[j00]+X1*gH[j10]+X2*gHa[j00]+X3*gHa[j10]+
	hq*(E0*f0+E1*f1+hq*(E2*fq0+E3*fq1));

      f0	=x0*(gHq [j10]-gHq [j00])+x1*gHaq [j00]+x2*gHaq [j10];
      f1	=x0*(gHq [j11]-gHq [j01])+x1*gHaq [j01]+x2*gHaq [j11];
      fq0	=x0*(gHqq[j10]-gHqq[j00])+x1*gHaqq[j00]+x2*gHaqq[j10];
      fq1	=x0*(gHqq[j11]-gHqq[j01])+x1*gHaqq[j01]+x2*gHaqq[j11];
      gha[k]	=x0*(gH[j10]-gH[j00])+x1*gHa[j00]+x2*gHa[j10]+
	hq*(E0*f0+E1*f1+hq*(E2*fq0+E3*fq1));

      F[k]	=X0*aF[i0]+X1*aF[i1]+X2*daF[i0]+X3*daF[i1];
      Fa[k]	=x0*(aF[i1]-aF[i0])+x1*daF[i0]+x2*daF[i1];
      gFa[k]	=X0*dgF[i0]+X1*dgF[i1]+X2*d2gF[i0]+X3*d2gF[i1];
      gFaa[k]	=x0*(dgF[i1]-dgF[i0])+x1*d2gF[i0]+x2*d2gF[i1];
      gFaa[k]	=a*gFaa[k]+gFa[k];
      gYa[k]	=X0*dgY[i0]+X1*dgY[i1]+X2*d2gY[i0]+X3*d2gY[i1];
      gYaa[k]	=x0*(dgY[i1]-dgY[i0])+x1*d2gY[i0]+x2*d2gY[i1];
      gYaa[k]	=a*gYaa[k]+gYa[k];
      T[k]	=X0*aT[i0]+X1*aT[i1]+X2*daT[i0]+X3*daT[i1];
      Ta[k]	=x0*(aT[i1]-aT[i0])+x1*daT[i0]+x2*daT[i1];
      P[k]	=X0*aP[i0]+X1*aP[i1]+X2*daP[i0]+X3*daP[i1];
      Pa[k]	=x0*(aP[i1]-aP[i0])+x1*daP[i0]+x2*daP[i1];
      gFa[k]	*=a;
      gYa[k]	*=a;
    }
    else{	/* a == 0. rq/a zq/a Bq/a */
      r[k]	=sr[0];
      ra[k]	=Y0*sra[j0]+Y1*sra[j1]+Y2*sraq[j0]+Y3*sraq[j1];
      rq[k]	=y0*(sra[j1]-sra[j0])+y1*sraq[j0]+y2*sraq[j1];
      z[k]	=sz[0];
      za[k]	=Y0*sza[j0]+Y1*sza[j1]+Y2*szaq[j0]+Y3*szaq[j1];
      zq[k]	=y0*(sza[j1]-sza[j0])+y1*szaq[j0]+y2*szaq[j1];
      B[k]	=aB[0];
      Ba[k]	=Y0*aBa[j0]+Y1*aBa[j1]+Y2*aBaq[j0]+Y3*aBaq[j1];
      Bq[k]	=y0*(aBa[j1]-aBa[j0])+y1*aBaq[j0]+y2*aBaq[j1];
      ghq[k]	=Y0*gHaq[j0]+Y1*gHaq[j1]+Y2*gHaqq[j0]+Y3*gHaqq[j1];
      gh[k]	=0.;
      gha[k]	=gHa[j0]
	+hq*(E0*gHaq[j0]+E1*gHaq[j1]+hq*(E2*gHaqq[j0]+E3*gHaqq[j1]));
      F[k]	=aF[0];
      Fa[k]	=0.;
      gFa[k]	=dgF[0];
      gFaa[k]	=dgF[0];
      gYa[k]	=dgY[0];
      gYaa[k]	=dgY[0];
      T[k]	=aT[0];
      Ta[k]	=0.;
      P[k]	=aP[0];
      Pa[k]	=0.;
    }
  }
  return(0);
}
#endif /*stg_Basic4User*/

/***************** Service routines ****************************/ 
#ifndef stg_Service4Hacker
int ESI4Hacker(ESI **lE, int i)
{
  if(i < 8 && sESI[i].Na != 0){
    *lE	=sESI+i;
  }
  else{
    *lE	=NULL;
    return(1);
  }
  return(0);
}

int ESIGetSize(int *na,int *np)
{
  *na	=Na;
  *np	=Np;
  return(0);
}

int ESIGetLink2Geometry(double **Xr,double **Xra,double **Xrq
		   ,double **Xz,double**Xza,double **Xzq)
{
  *Xr	=r;
  *Xra	=ra;
  *Xrq	=rq;
  *Xz	=z;
  *Xza	=za;
  *Xzq	=zq;
  return(0);
}
#endif /*stg_Service4Hacker*/

#ifndef stg_Service4User
int esigetdensity_(double *ne,double *dne,double *Te,double *Ti,double *a)
{
  int i0,i1;
  double t,s;
  double p1,p2,p3;

  s	=cbsp2ne/((*Te)+(*Ti));

  i0	=(*a-sa[0])*rha;
  if(i0 >= Na) i0=Na-1;
  i1	=i0+1;

  p1	=ha*dbsp[i0];
  p3	=ha*dbsp[i1];
  t	=bsp[i1]-bsp[i0];
  p2	=3.*t-2.*p1-p3;
  p3	+=p1-2.*t;
  t	=(*a-sa[i0])*rha;

  *ne	=(bsp[i0]+(p1+(p2+p3*t)*t)*t)*s;
  if(dne != NULL){
    *dne	=rha*(p1+(2.*p2+3.*p3*t)*t)*s;
  }
  return(0);
}

int ESI2Ipl(double *Ipl)
{
  int j,ji;

  *Ipl	=0.;
  ji	=Np1*Na;
  for(j=0; j < Np; j++){
    *Ipl	+=(srq[ji]*srq[ji]+szq[ji]*szq[ji])
      /(sr[ji]*(sra[ji]*szq[ji]-sza[ji]*srq[ji]));
    ji++;
  }
  *Ipl	*=-5.*dgY[Na]/Np;
  return(0);
}

int ESI2gm(double *gm,double *gma,double a)
{
  double x,X,xx,XX;
  double F,Fa,Y,Ya;

  if(a < 0.) a=-a;
  i0	=(a-sa[0])*rha;
  if(i0 >= Na) i0=Na-1;
  i1	=i0+1;
  x	=(a-sa[i0])*rha;
  X	=1.-x;

  XX	=X*X;
  xx	=x*x;
  XF	=XX*(3.-2.*X);
  XD	=XX*x*ha;
  xF	=xx*(3.-2.*x);
  xD	=-xx*X*ha;

  dX	=3.*x*X;
  dXD	=X-dX;
  dxD	=x-dX;
  dX	*=2.*rha;

  F	=XF*dgF[i0]+xF*dgF[i1]+XD*d2gF[i0]+xD*d2gF[i1];
  Fa	=dX*(dgF[i1]-dgF[i0])+dXD*d2gF[i0]+dxD*d2gF[i1];
  Y	=XF*dgY[i0]+xF*dgY[i1]+XD*d2gY[i0]+xD*d2gY[i1];
  Ya	=dX*(dgY[i1]-dgY[i0])+dXD*d2gY[i0]+dxD*d2gY[i1];
  x	=-1./F;
  *gm	=Y*x;
  if(gma != NULL) *gma=(Ya+(*gm)*Fa)*x;
  return(0);
}

int esiget2d_(int m, REAL *G, REAL *Ga, REAL *Gq, REAL *a0,REAL *gq0,int *n)
{
  int j0,j1,k;
  REAL x,X,xx,XX;
  REAL f0,f1,fq0,fq1;
  REAL a,q;
  double *sZ,*sZa,*sZq,*sZaq;
  
  switch(m){
  case 0:
    sZ	=sr;
    sZa	=sra;
    sZq	=srq;
    sZaq=sraq;
    break;
  case 1:
    sZ	=sz;
    sZa	=sza;
    sZq	=szq;
    sZaq=szaq;
    break;
  case 2:
    sZ	=aB;
    sZa	=aBa;
    sZq	=aBq;
    sZaq=aBaq;
    break;
  default:
    break;
  }

  for(k =0; k < *n; k++){
    a	=a0[k];
    q	=gq0[k];
    j1	=q*crgq0;
    if(q < 0.){
      j1--;
    }
    q	-=cgq0*j1;
    j0	=q*rhq;
    if(j0 >= Np){
      j0	-=Np;
      q	-=cgq0;
    }
    if(j0 < 0){
      j0	+=Np;
      q	+=cgq0;
    }
    j1	=j0+1;
    y	=(q-gq[j0])*rhq;
    Y	=1.-y;
   
    YY	=Y*Y;
    yy	=y*y;
    YF	=YY*(3.-2.*Y);
    YD	=YY*y*hq;
    yF	=yy*(3.-2.*y);
    yD	=-yy*Y*hq;
    
    dY	=3.*y*Y;
    dYD	=Y-dY;
    dyD	=y-dY;
    dY	*=2.*rhq;
    if(a < 0.){
      a	=-a;
    }
    i0	=(a-sa[0])*rha;
    if(a != 0.){
      if(i0 >= Na){
	i0	=Na-1;
      }
      i1	=i0+1;
      
      x	=(a-sa[i0])*rha;
      X	=1.-x;

      XX	=X*X;
      xx	=x*x;
      XF	=XX*(3.-2.*X);
      XD	=XX*x*ha;
      xF	=xx*(3.-2.*x);
      xD	=-xx*X*ha;
      
      dX	=3.*x*X;
      dXD	=X-dX;
      dxD	=x-dX;
      dX	*=2.*rha;
      
      j00	=Np1*i0+j0;
      j01	=j00+1;
      j10	=j00+Np1;
      j11	=j10+1;

      f0	=XF*sZ [j00]+xF*sZ [j10]+XD*sZa [j00]+xD*sZa [j10];
      f1	=XF*sZ [j01]+xF*sZ [j11]+XD*sZa [j01]+xD*sZa [j11];
      fq0	=XF*sZq[j00]+xF*sZq[j10]+XD*sZaq[j00]+xD*sZaq[j10];
      fq1	=XF*sZq[j01]+xF*sZq[j11]+XD*sZaq[j01]+xD*sZaq[j11];
      G[k]=YF*f0+yF*f1+YD*fq0+yD*fq1;
      Gq[k]=dY*(f1-f0)+dYD*fq0+dyD*fq1;
      f0	=dX*(sZ [j10]-sZ [j00])+dXD*sZa [j00]+dxD*sZa [j10];
      f1	=dX*(sZ [j11]-sZ [j01])+dXD*sZa [j01]+dxD*sZa [j11];
      fq0	=dX*(sZq[j10]-sZq[j00])+dXD*sZaq[j00]+dxD*sZaq[j10];
      fq1	=dX*(sZq[j11]-sZq[j01])+dXD*sZaq[j01]+dxD*sZaq[j11];
      Ga[k]=YF*f0+yF*f1+YD*fq0+yD*fq1;
    }
    else{	/* a == 0. */
      G[k]	=sZ[0];
      Ga[k]	=YF*sZa[j0]+yF*sZa[j1]+YD*sZaq[j0]+yD*sZaq[j1];
      Gq[k]	=dY*(sZa[j1]-sZa[j0])+dYD*sZaq[j0]+dyD*sZaq[j1];
    }
  }
  return(0);
}

int esigetrzb_(REAL *R, REAL *Z, REAL *Bm,REAL *a0,REAL *q0,int *n)
{
  int j0,j1,k;
  REAL x,X,xx,XX;
  REAL y,Y,yy,YY;
  REAL f0,f1,fq0,fq1;
  REAL a,q;

  for(k =0; k < *n; k++){
    a	=a0[k];
    q	=q0[k];
    
    if(a < 0.) a=-a;
    i0	=(a-sa[0])*rha;
    if(i0 >= Na) i0=Na-1;
    i1	=i0+1;
    
    x	=(a-sa[i0])*rha;
    X	=1.-x;
    
    XX	=X*X;
    xx	=x*x;
    XF	=XX*(3.-2.*X);
    XD	=XX*x*ha;
    xF	=xx*(3.-2.*x);
    xD	=-xx*X*ha;
    
    dX	=3.*x*X;
    dXD	=X-dX;
    dxD	=x-dX;
    dX	*=2.*rha;
    
    j1	=q*crgq0;
    if(q < 0.) j1--;
    q	-=cgq0*j1;
    j0	=q*rhq;
    if(j0 >= Np){
      j0	-=Np;
      q	-=cgq0;
    }
    if(j0 < 0){
      j0	+=Np;
      q	+=cgq0;
    }
    j1	=j0+1;
    
    Y	=(gq[j1]-q)*rhq;
    y	=1.-Y;

    YY	=Y*Y;
    yy	=y*y;
    YF	=YY*(3.-2.*Y);
    YD	=YY*y*hq;
    yF	=yy*(3.-2.*y);
    yD	=-yy*Y*hq;
    
    dY	=3.*y*Y;
    dYD	=Y-dY;
    dyD	=y-dY;
    dY	*=2.*rhq;
    
    j00	=Np1*i0+j0;
    j01	=j00+1;
    j10	=j00+Np1;
    j11	=j10+1;
    
    f0	=XF*sr [j00]+xF*sr [j10]+XD*sra [j00]+xD*sra [j10];
    f1	=XF*sr [j01]+xF*sr [j11]+XD*sra [j01]+xD*sra [j11];
    fq0	=XF*srq[j00]+xF*srq[j10]+XD*sraq[j00]+xD*sraq[j10];
    fq1	=XF*srq[j01]+xF*srq[j11]+XD*sraq[j01]+xD*sraq[j11];
    R[k]=YF*f0+yF*f1+YD*fq0+yD*fq1;
    
    f0	=XF*sz [j00]+xF*sz [j10]+XD*sza [j00]+xD*sza [j10];
    f1	=XF*sz [j01]+xF*sz [j11]+XD*sza [j01]+xD*sza [j11];
    fq0	=XF*szq[j00]+xF*szq[j10]+XD*szaq[j00]+xD*szaq[j10];
    fq1	=XF*szq[j01]+xF*szq[j11]+XD*szaq[j01]+xD*szaq[j11];
    Z[k]=YF*f0+yF*f1+YD*fq0+yD*fq1;
    
    if(Bm != NULL){
      f0	=XF*aB [j00]+xF*aB [j10]+XD*aBa [j00]+xD*aBa [j10];
      f1	=XF*aB [j01]+xF*aB [j11]+XD*aBa [j01]+xD*aBa [j11];
      fq0	=XF*aBq[j00]+xF*aBq[j10]+XD*aBaq[j00]+xD*aBaq[j10];
      fq1	=XF*aBq[j01]+xF*aBq[j11]+XD*aBaq[j01]+xD*aBaq[j11];
      Bm[k]=YF*f0+yF*f1+YD*fq0+yD*fq1;
    }
  }
  return(0);
}

int esiGetgH(double *gH0,double *gH0q,double *a0,double *gq0,int n)
{
  /* Exception: at a=0 the routine returns:
     dr_q/a, dz_q/a, dB_q/a, dgF/a, dgY/a
     rather than
     dr_q=0, dz_q=0, dB_q=0, dgF=0, dgY=0 */
  int j0,j1,k;
  double x,X,xx,XX;
  double f0,f1,fq0,fq1;
  double a,q;

  for(k =0; k < n; k++){
    a	=a0[k];
    q	=gq0[k];
    j1	=q*crgq0;
    if(q < 0.) j1--;
    q	-=cgq0*j1;
    j0	=q*rhq;
    if(j0 >= Np){
      j0	-=Np;
      q	-=cgq0;
    }
    if(j0 < 0){
      j0	+=Np;
      q	+=cgq0;
    }
    j1	=j0+1;
    y	=(q-gq[j0])*rhq;
    Y	=1.-y;
   
    YY	=Y*Y;
    yy	=y*y;
    YF	=YY*(3.-2.*Y);
    YD	=YY*y*hq;
    yF	=yy*(3.-2.*y);
    yD	=-yy*Y*hq;
    
    dY	=3.*y*Y;
    dYD	=Y-dY;
    dyD	=y-dY;
    dY	*=2.*rhq;
    if(a < 0.) a=-a;
    i0	=(a-sa[0])*rha;
    if(a != 0.){
      if(i0 >= Na) i0=Na-1;
      i1	=i0+1;
      x	=(a-sa[i0])*rha;
      X	=1.-x;

      XX	=X*X;
      xx	=x*x;
      XF	=XX*(3.-2.*X);
      XD	=XX*x*ha;
      xF	=xx*(3.-2.*x);
      xD	=-xx*X*ha;
      
      dX	=3.*x*X;
      dXD	=X-dX;
      dxD	=x-dX;
      dX	*=2.*rha;
      
      j00	=Np1*i0+j0;
      j01	=j00+1;
      j10	=j00+Np1;
      j11	=j10+1;

      /* gh,gh'_a,gh'_q */
      f0	=XF*gHq [j00]+xF*gHq [j10]+XD*gHaq [j00]+xD*gHaq [j10];
      f1	=XF*gHq [j01]+xF*gHq [j11]+XD*gHaq [j01]+xD*gHaq [j11];
      fq0	=XF*gHqq[j00]+xF*gHqq[j10]+XD*gHaqq[j00]+xD*gHaqq[j10];
      fq1	=XF*gHqq[j01]+xF*gHqq[j11]+XD*gHaqq[j01]+xD*gHaqq[j11];
      if(gH0q != NULL)  gH0q[k]=YF*f0+yF*f1+YD*fq0+yD*fq1;

      gH0[k]	=XF*gH[j00]+xF*gH[j10]+XD*gHa[j00]+xD*gHa[j10]+
	hq*0.5*((1.+YY*(YY-2.*Y))*f0+yy*(2.*y-yy)*f1
		+cr6*hq*((1.-YY*(YY+4.*Y*y))*fq0-yy*(4.*y*Y+yy)*fq1));
#ifdef H
      f0	=dX*(gHq [j10]-gHq [j00])+dXD*gHaq [j00]+dxD*gHaq [j10];
      f1	=dX*(gHq [j11]-gHq [j01])+dXD*gHaq [j01]+dxD*gHaq [j11];
      fq0	=dX*(gHqq[j10]-gHqq[j00])+dXD*gHaqq[j00]+dxD*gHaqq[j10];
      fq1	=dX*(gHqq[j11]-gHqq[j01])+dXD*gHaqq[j01]+dxD*gHaqq[j11];
      gH0a[k]	=dX*(gH[j10]-gH[j00])+dXD*gHa[j00]+dxD*gHa[j10]+
	hq*0.5*((1.+YY*(YY-2.*Y))*f0+yy*(2.*y-yy)*f1
		+cr6*hq*((1.-YY*(YY+4.*Y*y))*fq0-yy*(4.*y*Y+yy)*fq1));
#endif
    }
    else{	/* a == 0. */
      if(gH0q !=NULL)gH0q[k]=YF*gHaq[j0]+yF*gHaq[j1]+YD*gHaqq[j0]+yD*gHaqq[j1];
      gH0[k]	=0.;
#ifdef H
      gH0a[k]	=gHa[j0]+hq*0.5*((1.+YY*(YY-2.*Y))*gHaq[j0]
				 +yy*(2.*y-yy)*gHaq[j1]
				 +cr6*hq*((1.-YY*(YY+4.*Y*y))*gHaqq[j0]
					  -yy*(4.*y*Y+yy)*gHaqq[j1]));
#endif
    }
  }
  return(0);
}

int gcmotion_(REAL *dXgr,REAL *dXa,REAL *dXgq,REAL *dXgf
	      ,REAL *Xgr,REAL *Xa,REAL *Xgq,REAL *Xgm,int*n)
{
  int j0,j1,k;
  REAL r,ra,rq,za,zq,B,Ba,Bq,F,gYa,T,P;
  REAL a,a11,a12,a21,a22;

  double t,pt;
  double X0,X1,X2,X3;
  double x0,x1,x2;
  double Y0,Y1,Y2,Y3;
  double y0,y1,y2;

  REAL f0,f1,fq0,fq1;
  REAL D;

  for(k=0; k < *n; k++){
    t	=Xgq[k];
    j1	=t*crgq0;
    if(t < 0.) j1--;
    t	-=cgq0*j1;
    j0	=t*rhq;
    if(j0 >= Np){
      j0	-=Np;
      t	-=cgq0;
    }
    if(j0 < 0){
      j0	+=Np;
      t	+=cgq0;
    }
    j1	=j0+1;
    t	=(t-gq[j0])*rhq;

    pt	=t*t;
    y1	=2.-2.*t;
    Y1	=(1.+y1)*pt;
    Y0	=1.-Y1;
    y0	=t-pt;
    Y3	=-y0*t*hq;
    Y2	=Y3+y0*hq;
    y0	*=6.*rhq;
    y2	=pt-y1*t;
    y1	+=y2-1.;

    a	=Xa[k];
    if(a < 1e-100) a=0.;
    if(a < 0.) a=-a;
    i0	=(a-sa[0])*rha;
    if(a != 0.){
      if(i0 >= Na) i0=Na-1;
      i1	=i0+1;
      t		=(a-sa[i0])*rha;

      pt	=t*t;
      x1	=2.-2.*t;
      X1	=(1.+x1)*pt;
      X0	=1.-X1;
      x0	=t-pt;
      X3	=-x0*t*ha;
      X2	=X3+x0*ha;
      x0	*=6.*rha;
      x2	=pt-x1*t;
      x1	+=x2-1.;
      
      j00	=Np1*i0+j0;
      j01	=j00+1;
      j10	=j00+Np1;
      j11	=j10+1;
      
      /* r,r'_a,r'_q */
      f0	=X0*sr [j00]+X1*sr [j10]+X2*sra [j00]+X3*sra [j10];
      f1	=X0*sr [j01]+X1*sr [j11]+X2*sra [j01]+X3*sra [j11];
      fq0	=X0*srq[j00]+X1*srq[j10]+X2*sraq[j00]+X3*sraq[j10];
      fq1	=X0*srq[j01]+X1*srq[j11]+X2*sraq[j01]+X3*sraq[j11];
      r		=Y0*f0+Y1*f1+Y2*fq0+Y3*fq1;
      rq	=y0*(f1-f0)+y1*fq0+y2*fq1;
      f0	=x0*(sr [j10]-sr [j00])+x1*sra [j00]+x2*sra [j10];
      f1	=x0*(sr [j11]-sr [j01])+x1*sra [j01]+x2*sra [j11];
      fq0	=x0*(srq[j10]-srq[j00])+x1*sraq[j00]+x2*sraq[j10];
      fq1	=x0*(srq[j11]-srq[j01])+x1*sraq[j01]+x2*sraq[j11];
      ra	=Y0*f0+Y1*f1+Y2*fq0+Y3*fq1;
      
      /* z,z'_a,z'_q */
      f0	=X0*sz [j00]+X1*sz [j10]+X2*sza [j00]+X3*sza [j10];
      f1	=X0*sz [j01]+X1*sz [j11]+X2*sza [j01]+X3*sza [j11];
      fq0	=X0*szq[j00]+X1*szq[j10]+X2*szaq[j00]+X3*szaq[j10];
      fq1	=X0*szq[j01]+X1*szq[j11]+X2*szaq[j01]+X3*szaq[j11];
      zq	=y0*(f1-f0)+y1*fq0+y2*fq1;
      f0	=x0*(sz [j10]-sz [j00])+x1*sza [j00]+x2*sza [j10];
      f1	=x0*(sz [j11]-sz [j01])+x1*sza [j01]+x2*sza [j11];
      fq0	=x0*(szq[j10]-szq[j00])+x1*szaq[j00]+x2*szaq[j10];
      fq1	=x0*(szq[j11]-szq[j01])+x1*szaq[j01]+x2*szaq[j11];
      za	=Y0*f0+Y1*f1+Y2*fq0+Y3*fq1;
      
      /* B,B'_a,B'_q */
      f0	=X0*aB [j00]+X1*aB [j10]+X2*aBa [j00]+X3*aBa [j10];
      f1	=X0*aB [j01]+X1*aB [j11]+X2*aBa [j01]+X3*aBa [j11];
      fq0	=X0*aBq[j00]+X1*aBq[j10]+X2*aBaq[j00]+X3*aBaq[j10];
      fq1	=X0*aBq[j01]+X1*aBq[j11]+X2*aBaq[j01]+X3*aBaq[j11];
      B		=Y0*f0+Y1*f1+Y2*fq0+Y3*fq1;
      Bq	=y0*(f1-f0)+y1*fq0+y2*fq1;
      f0	=x0*(aB [j10]-aB [j00])+x1*aBa [j00]+x2*aBa [j10];
      f1	=x0*(aB [j11]-aB [j01])+x1*aBa [j01]+x2*aBa [j11];
      fq0	=x0*(aBq[j10]-aBq[j00])+x1*aBaq[j00]+x2*aBaq[j10];
      fq1	=x0*(aBq[j11]-aBq[j01])+x1*aBaq[j01]+x2*aBaq[j11];
      Ba	=Y0*f0+Y1*f1+Y2*fq0+Y3*fq1;

      F		=X0*aF[i0]+X1*aF[i1]+X2*daF[i0]+X3*daF[i1];
      gYa	=X0*dgY[i0]+X1*dgY[i1]+X2*d2gY[i0]+X3*d2gY[i1];
      T		=X0*aT[i0]+X1*aT[i1]+X2*daT[i0]+X3*daT[i1];
      P		=X0*aP[i0]+X1*aP[i1]+X2*daP[i0]+X3*daP[i1];
      gYa	*=a;
    }
    else{	/* a == 0. */
      r		=sr[0];
      ra	=Y0*sra[j0]+Y1*sra[j1]+Y2*sraq[j0]+Y3*sraq[j1];
      rq	=y0*(sra[j1]-sra[j0])+y1*sraq[j0]+y2*sraq[j1];
      za	=Y0*sza[j0]+Y1*sza[j1]+Y2*szaq[j0]+Y3*szaq[j1];
      zq	=y0*(sza[j1]-sza[j0])+y1*szaq[j0]+y2*szaq[j1];
      B		=aB[0];
      Ba	=Y0*aBa[j0]+Y1*aBa[j1]+Y2*aBaq[j0]+Y3*aBaq[j1];
      Bq	=y0*(aBa[j1]-aBa[j0])+y1*aBaq[j0]+y2*aBaq[j1];
      F		=aF[0];
      gYa	=dgY[0];
      T		=aT[0];
      P		=aP[0];
      
      T		*=Xgr[k];
      D		=za*rq-zq*ra;
      a21	=1./(D*((F+T)/r+Xgr[k]*r*P));
#ifdef H
      r  =cgq0*Xgr[k]*B;
      B	*=r;
      r	=r*Xgr[k]+cgq0*Xgm[k];
#endif
      r  =Xgr[k]*B;
      B	*=r;
      r	=r*Xgr[k]+Xgm[k];

      Ba	*=r;
      Bq	*=-r;
      dXgr[k]	=0.;
      Bq	*=a21;
      Ba	*=a21;
      dXgf[k]	=B/F;
      T	=cos(Xgq[k]);
      P	=sin(Xgq[k]);
      dXa[k]	=Bq*T-Ba*P;
      dXgq[k]	=Bq*P+Ba*T;
    }
    a21	=F;
    T	*=Xgr[k];
    D	=za*rq-zq*ra;
    a12	=D*((F+T)/r+Xgr[k]*r*P);
    D	=-gYa/(r*D);
    a11	=(rq*rq+zq*zq)*D;
    D	*=ra*rq+za*zq;
    a22	=gYa*(1.+T/F);
    r	=1./(a11*a22-a12*a21);
    a11	*=r;
    a12	*=r;
    a21	*=r;
    a22	*=r;
    D	*=r;
#ifdef H
    r  =Xgr[k]*B*cgq0;
    B	*=r;
    r	=r*Xgr[k]+Xgm[k]*cgq0;
#endif
    r  =Xgr[k]*B;
    B	*=r;
    r	=r*Xgr[k]+Xgm[k];
    Ba	*=r;
    Bq	*=-r;
    dXgr[k]	=a22*Bq;
    dXgf[k]	=-a12*B+a11*Ba+D*Bq;
    dXa[k]	=-a21*Bq;
    dXgq[k]	=a22*B-a21*Ba;
    if(sw != NULL && sw[k]){
      Bq	=a*dXgq[k];
      Ba	=dXa[k];
      T	=cos(Xgq[k]);
      P	=sin(Xgq[k]);
      dXa[k]	=Ba*T-Bq*P;
      dXgq[k]	=Ba*P+Bq*T;
    }
  }
  return(0);
}

int MFieldLines(REAL *dXgq,REAL *dXgf,REAL *Xa,REAL *Xgq,int n) 
{
  int j0,j1,k;
  REAL r,ra,rq,za,zq,B,Ba,Bq,F,gYa;
  REAL a,q;
  REAL x,X,xx,XX;
  REAL f0,f1,fq0,fq1;
  REAL y,Y,yy,YY;
  REAL D;

  for(k=0; k < n; k++){
    a	=Xa[k];
    if(a < 1e-100){
      a	=0.;
    }
    q	=Xgq[k];

    j1	=q*crgq0;
    if(q < 0.){
      j1--;
    }
    q	-=cgq0*j1;
    j0	=q*rhq;
    if(j0 >= Np){
      j0	-=Np;
      q	-=cgq0;
    }
    if(j0 < 0){
      j0	+=Np;
      q	+=cgq0;
    }
    j1	=j0+1;
    y	=(q-gq[j0])*rhq;

    Y	=1.-y;
    YY	=Y*Y;
    yy	=y*y;
    YF	=YY*(3.-2.*Y);
    YD	=YY*y*hq;
    yF	=yy*(3.-2.*y);
    yD	=-yy*Y*hq;
    dY	=3.*y*Y;
    dYD	=Y-dY;
    dyD	=y-dY;
    dY	*=2.*rhq;

    if(a < 0.){
      a	=-a;
    }
    i0	=(a-sa[0])*rha;
    if(i0 >= Na){
      i0	=Na-1;
    }
    i1	=i0+1;
    x		=(a-sa[i0])*rha;
    X		=1.-x;
    XX	=X*X;
    xx	=x*x;
    XF	=XX*(3.-2.*X);
    XD	=XX*x*ha;
    xF	=xx*(3.-2.*x);
    xD	=-xx*X*ha;
    
    dX	=3.*x*X;
    dXD	=X-dX;
    dxD	=x-dX;
    dX	*=2.*rha;
    
    j00	=Np1*i0+j0;
    j01	=j00+1;
    j10	=j00+Np1;
    j11	=j10+1;
    
    /* r,r'_a,r'_q */
    f0	=XF*sr [j00]+xF*sr [j10]+XD*sra [j00]+xD*sra [j10];
    f1	=XF*sr [j01]+xF*sr [j11]+XD*sra [j01]+xD*sra [j11];
    fq0	=XF*srq[j00]+xF*srq[j10]+XD*sraq[j00]+xD*sraq[j10];
    fq1	=XF*srq[j01]+xF*srq[j11]+XD*sraq[j01]+xD*sraq[j11];
    r		=YF*f0+yF*f1+YD*fq0+yD*fq1;
    rq	=dY*(f1-f0)+dYD*fq0+dyD*fq1;
    f0	=dX*(sr [j10]-sr [j00])+dXD*sra [j00]+dxD*sra [j10];
    f1	=dX*(sr [j11]-sr [j01])+dXD*sra [j01]+dxD*sra [j11];
    fq0	=dX*(srq[j10]-srq[j00])+dXD*sraq[j00]+dxD*sraq[j10];
    fq1	=dX*(srq[j11]-srq[j01])+dXD*sraq[j01]+dxD*sraq[j11];
    ra	=YF*f0+yF*f1+YD*fq0+yD*fq1;
    
    /* z,z'_a,z'_q */
    f0	=XF*sz [j00]+xF*sz [j10]+XD*sza [j00]+xD*sza [j10];
    f1	=XF*sz [j01]+xF*sz [j11]+XD*sza [j01]+xD*sza [j11];
    fq0	=XF*szq[j00]+xF*szq[j10]+XD*szaq[j00]+xD*szaq[j10];
    fq1	=XF*szq[j01]+xF*szq[j11]+XD*szaq[j01]+xD*szaq[j11];
    zq	=dY*(f1-f0)+dYD*fq0+dyD*fq1;
    f0	=dX*(sz [j10]-sz [j00])+dXD*sza [j00]+dxD*sza [j10];
    f1	=dX*(sz [j11]-sz [j01])+dXD*sza [j01]+dxD*sza [j11];
    fq0	=dX*(szq[j10]-szq[j00])+dXD*szaq[j00]+dxD*szaq[j10];
    fq1	=dX*(szq[j11]-szq[j01])+dXD*szaq[j01]+dxD*szaq[j11];
    za	=YF*f0+yF*f1+YD*fq0+yD*fq1;
    
    /* B,B'_a,B'_q */
    f0	=XF*aB [j00]+xF*aB [j10]+XD*aBa [j00]+xD*aBa [j10];
    f1	=XF*aB [j01]+xF*aB [j11]+XD*aBa [j01]+xD*aBa [j11];
    fq0	=XF*aBq[j00]+xF*aBq[j10]+XD*aBaq[j00]+xD*aBaq[j10];
    fq1	=XF*aBq[j01]+xF*aBq[j11]+XD*aBaq[j01]+xD*aBaq[j11];
    B		=YF*f0+yF*f1+YD*fq0+yD*fq1;
    Bq	=dY*(f1-f0)+dYD*fq0+dyD*fq1;
    f0	=dX*(aB [j10]-aB [j00])+dXD*aBa [j00]+dxD*aBa [j10];
    f1	=dX*(aB [j11]-aB [j01])+dXD*aBa [j01]+dxD*aBa [j11];
    fq0	=dX*(aBq[j10]-aBq[j00])+dXD*aBaq[j00]+dxD*aBaq[j10];
    fq1	=dX*(aBq[j11]-aBq[j01])+dXD*aBaq[j01]+dxD*aBaq[j11];
    Ba	=YF*f0+yF*f1+YD*fq0+yD*fq1;
    
    gYa	=XF*dgY[i0]+xF*dgY[i1]+XD*d2gY[i0]+xD*d2gY[i1];
    gYa	*=a;

    D	=za*rq-zq*ra;
    dXgq[k]	=-gYa/(r*D*B);
    dXgf[k]	=F/(r*r*B);
  }
  return(0);
}

void esigetpressure_(double *p,double *dp,double *a, int *n)
{
  int k,i;
  double p0[Na1],dp0,dp1,d2p0,d2p1,x,X,p3x,p3X,XF,XD,xF,xD;

  p0[0]	=0.;
  dp1	=0.;
  d2p1	=aP[0]*dgY[0];
  for(i=1,k=0; i < Na1; i++,k++){
    dp0	=dp1;
    d2p0=d2p1;
    dp1	=aP[i]*dgY[i]*sa[i];
    d2p1=aP[i]*dgY[i]+(daP[i]*dgY[i]+aP[i]*d2gY[i])*sa[i];
    p0[i]=p0[k]+0.5*ha*((dp0+dp1)+ha*(d2p0-d2p1)/6.);
  }
  for(i=0; i < Na1; i++){
    p0[i]	-=p0[Na];
  }
  for(k=0; k < *n; k++){
    i	=(a[k]-sa[0])*rha;
    if(i >= Na){
      i	=Na-1;
    }
    x	=(a[k]-sa[i])/ha;
    X	=1.-x;
    XF	=X*X*(3.-2.*X);
    XD	=X*X*x*ha;
    xF	=x*x*(3.-2.*x);
    xD	=-x*x*X*ha;
    p3x	=x*x*x;
    p3X	=X*X*X;
    p[k]=p0[i]+0.5*ha*((1.+p3X*(X-2.))*aP[i]+p3x*(2.-x)*aP[i+1]
		       +ha*((1.-p3X*(X+4.*x))*daP[i]
			    -p3x*(4.*X+x)*daP[i+1])/6.);
    dp[k]	=(XF*aP[i]+xF*aP[i+1]+XD*daP[i]+xD*daP[i+1])*
      (XF*dgY[i]+xF*dgY[i+1]+XD*d2gY[i]+xD*d2gY[i+1])*a[k];
  }
  return;
}

int ESI2Geom(double *R,double *Ra,double *Rq
	       ,double *Z,double *Za,double *Zq
	       , double *a0,double *gq0,int n)
{
  /* Exception: at a=0 the routine returns:
     dr_q/a, dz_q/a, dB_q/a, dgF/a, dgY/a
     rather than
     dr_q=0, dz_q=0, dB_q=0, dgF=0, dgY=0 */
  int j0,j1,k;
  double x,X,xx,XX;
  double f0,f1,fq0,fq1;
  double a,q;

  for(k =0; k < n; k++){
    a	=a0[k];
    q	=gq0[k];
    j1	=q*crgq0;
    if(q < 0.) j1--;
    q	-=cgq0*j1;
    j0	=q*rhq;
    while(j0 >= Np){
      j0	-=Np;
      q	-=cgq0;
    }
    while(j0 < 0){
      j0	+=Np;
      q	+=cgq0;
    }
    j1	=j0+1;
    y	=(q-gq[j0])*rhq;
    Y	=1.-y;
   
    YY	=Y*Y;
    yy	=y*y;
    YF	=YY*(3.-2.*Y);
    YD	=YY*y*hq;
    yF	=yy*(3.-2.*y);
    yD	=-yy*Y*hq;
    
    dY	=3.*y*Y;
    dYD	=Y-dY;
    dyD	=y-dY;
    dY	*=2.*rhq;
    if(a < 0.) a=-a;
    i0	=(a-sa[0])*rha;
    if(a != 0.){
      if(i0 >= Na) i0=Na-1;
      i1	=i0+1;
      x	=(a-sa[i0])*rha;
      X	=1.-x;
      XX	=X*X;
      xx	=x*x;
      XF	=XX*(3.-2.*X);
      XD	=XX*x*ha;
      xF	=xx*(3.-2.*x);
      xD	=-xx*X*ha;
      
      dX	=3.*x*X;
      dXD	=X-dX;
      dxD	=x-dX;
      dX	*=2.*rha;
      j00	=Np1*i0+j0;
      j01	=j00+1;
      j10	=j00+Np1;
      j11	=j10+1;
      
      /* r,r'_a,r'_q */
      f0	=XF*sr [j00]+xF*sr [j10]+XD*sra [j00]+xD*sra [j10];
      f1	=XF*sr [j01]+xF*sr [j11]+XD*sra [j01]+xD*sra [j11];
      fq0	=XF*srq[j00]+xF*srq[j10]+XD*sraq[j00]+xD*sraq[j10];
      fq1	=XF*srq[j01]+xF*srq[j11]+XD*sraq[j01]+xD*sraq[j11];
      R[k]	=YF*f0+yF*f1+YD*fq0+yD*fq1;
      Rq[k]	=dY*(f1-f0)+dYD*fq0+dyD*fq1;
      f0	=dX*(sr [j10]-sr [j00])+dXD*sra [j00]+dxD*sra [j10];
      f1	=dX*(sr [j11]-sr [j01])+dXD*sra [j01]+dxD*sra [j11];
      fq0	=dX*(srq[j10]-srq[j00])+dXD*sraq[j00]+dxD*sraq[j10];
      fq1	=dX*(srq[j11]-srq[j01])+dXD*sraq[j01]+dxD*sraq[j11];
      Ra[k]	=YF*f0+yF*f1+YD*fq0+yD*fq1;
     
      /* z,z'_a,z'_q */
      f0	=XF*sz [j00]+xF*sz [j10]+XD*sza [j00]+xD*sza [j10];
      f1	=XF*sz [j01]+xF*sz [j11]+XD*sza [j01]+xD*sza [j11];
      fq0	=XF*szq[j00]+xF*szq[j10]+XD*szaq[j00]+xD*szaq[j10];
      fq1	=XF*szq[j01]+xF*szq[j11]+XD*szaq[j01]+xD*szaq[j11];
      Z[k]	=YF*f0+yF*f1+YD*fq0+yD*fq1;
      Zq[k]	=dY*(f1-f0)+dYD*fq0+dyD*fq1;
      f0	=dX*(sz [j10]-sz [j00])+dXD*sza [j00]+dxD*sza [j10];
      f1	=dX*(sz [j11]-sz [j01])+dXD*sza [j01]+dxD*sza [j11];
      fq0	=dX*(szq[j10]-szq[j00])+dXD*szaq[j00]+dxD*szaq[j10];
      fq1	=dX*(szq[j11]-szq[j01])+dXD*szaq[j01]+dxD*szaq[j11];
      Za[k]	=YF*f0+yF*f1+YD*fq0+yD*fq1;
    }
    else{	/* a == 0. */
      R[k]	=sr[0];
      Ra[k]	=YF*sra[j0]+yF*sra[j1]+YD*sraq[j0]+yD*sraq[j1];
      Rq[k]	=dY*(sra[j1]-sra[j0])+dYD*sraq[j0]+dyD*sraq[j1];
      Z[k]	=sz[0];
      Za[k]	=YF*sza[j0]+yF*sza[j1]+YD*szaq[j0]+yD*szaq[j1];
      Zq[k]	=dY*(sza[j1]-sza[j0])+dYD*szaq[j0]+dyD*szaq[j1];
    }
  }
  return(0);
}
#endif /*stg_Service4User*/

#ifndef stg_Junk
/***************** Junc routines ****************************/
int ESICopy(double *esr,double *esz,double *eaB,double *egH,int na1,int np1)
{
  int i,n;
  double *esra,*esrt,*esrat,*esza,*eszt,*eszat,*eaBa,*eaBt,*eaBat
    ,*egHa,*egHt,*egHat;

  n	=np1*na1;

  esra	=esr	+n;
  esrt	=esra	+n;
  esrat	=esrt	+n;
  esza	=esz	+n;
  eszt	=esza	+n;
  eszat	=eszt	+n;
  eaBa	=eaB	+n;
  eaBt	=eaBa	+n;
  eaBat	=eaBt	+n;
  egHa	=egH	+n;
  egHt	=egHa	+n;
  egHat	=egHt	+n;
  for(i=0; i < n; i++){
    esr[i]	=sr[i];
    esra[i]	=sra[i];
    esrt[i]	=-srq[i];
    esrat[i]	=-sraq[i];
    esz[i]	=sz[i];
    esza[i]	=sza[i];
    eszt[i]	=-szq[i];
    eszat[i]	=-szaq[i];
    eaB[i]	=aB[i];
    eaBa[i]	=aBa[i];
    eaBt[i]	=-aBq[i];
    eaBat[i]	=-aBaq[i];
    egH[i]	=-gH[i];
    egHa[i]	=-gHa[i];
    egHt[i]	=gHq[i];
    egHat[i]	=gHaq[i];

    egH[i]	=gHq[i];
    egHa[i]	=gHaq[i];
    egHt[i]	=-gHqq[i];
    egHat[i]	=-gHaqq[i];
  }
  return(0);
}
#endif /*stg_Junk*/

#endif/*mzl_ESI*/
